"""Utility functions for web interactions."""
import asyncio
from functools import wraps
import inspect
import logging
import time
from typing import (
    Any,
    Awaitable,
    Callable,
    Final,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

import httpx
from httpx._client import USE_CLIENT_DEFAULT, UseClientDefault
from httpx._types import (
    AuthTypes,
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestFiles,
    TimeoutTypes,
    URLTypes,
)
import requests
from requests import Response

_logger = logging.getLogger(__name__)

_RETRY_STATUS_CODES: Final = (404, 500, 502, 503, 504)
_DEFAULT_TIMEOUT: Final = 20.0
_DEFAULT_MAX_RETRIES: Final = 3
_DEFAULT_BACKOFF_FACTOR: Final = 1

# These should be replaced with ParamSpec versions once
# https://github.com/python/mypy/issues/11855 is resolved
_SYNC_F = TypeVar("_SYNC_F", bound=Callable[..., Response])
_ASYNC_F = TypeVar("_ASYNC_F", bound=Callable[..., Awaitable[httpx.Response]])
_F = Union[_SYNC_F, _ASYNC_F]


@overload
def _auto_retry_request(
    original_req_func: _SYNC_F,
    *,
    max_retries: int = _DEFAULT_MAX_RETRIES,
    backoff_factor: int = _DEFAULT_BACKOFF_FACTOR,
) -> _SYNC_F:
    """Applies automatic retries to HTTP requests when encountering specific errors."""
    ...


@overload
def _auto_retry_request(
    original_req_func: _ASYNC_F,
    *,
    max_retries: int = _DEFAULT_MAX_RETRIES,
    backoff_factor: int = _DEFAULT_BACKOFF_FACTOR,
) -> _ASYNC_F:
    """Applies automatic retries to HTTP requests when encountering specific errors."""
    ...


@overload
def _auto_retry_request(
    original_req_func: None = None,
    *,
    max_retries: int = _DEFAULT_MAX_RETRIES,
    backoff_factor: int = _DEFAULT_BACKOFF_FACTOR,
) -> Callable[[_F], _F]:
    """Applies automatic retries to HTTP requests when encountering specific errors."""
    ...


def _auto_retry_request(
    original_req_func: Optional[_F] = None,
    *,
    max_retries: int = _DEFAULT_MAX_RETRIES,
    backoff_factor: int = _DEFAULT_BACKOFF_FACTOR,
) -> Union[_F, Callable[[_F], _F]]:
    """Applies automatic retries to HTTP requests when encountering specific errors.

    Wraps the target `requests` call in a retry mechanism which will reattempt
    the call if:
        - A connection error occurs
        - A retryable HTTP error response is received

    Utilises an exponential backoff to avoid flooding the request and to give
    time for the issue to resolve itself.

    Can be used as either an argumentless decorator (@_auto_retry_request) or a
    decorator with args (@_auto_retry_request(...)).
    """

    def _decorate(req_func: _F) -> _F:
        """Apply decoration to target request function."""
        if inspect.iscoroutinefunction(req_func):
            return _get_async_wrapper(req_func, max_retries, backoff_factor)
        else:
            return _get_sync_wrapper(req_func, max_retries, backoff_factor)

    if original_req_func:
        # Was used as @_auto_retry_dec (or called directly).
        # original_req_func was passed in through the decorator machinery so just
        # wrap and return.
        return _decorate(original_req_func)
    else:
        # Was used as @_auto_retry_dec(**kwargs).
        # original_req_func not yet passed in so need to return a decorator function
        # to allow the decorator machinery to pass it in.
        return _decorate


def _get_sync_wrapper(
    req_func: _SYNC_F, max_retries: int, backoff_factor: int
) -> _SYNC_F:
    @wraps(req_func)
    def _wrapped_sync_req_func(*args: Any, **kwargs: Any) -> Response:
        """Wraps target request function in retry capability.

        Adds automatic retry, backoff, and logging.
        """
        # Set default timeout if one not provided
        timeout = kwargs.get("timeout", None)
        if timeout is None:
            logging.debug(
                f"No request timeout provided,"
                f" setting to default timeout ({_DEFAULT_TIMEOUT}s)"
            )
            kwargs["timeout"] = _DEFAULT_TIMEOUT

        retry_count = 0

        while retry_count <= max_retries:
            final_retry = retry_count == max_retries

            # Attempt to make wrapped request and handle if it doesn't work
            # as expected
            try:
                resp: Response = req_func(*args, **kwargs)

                # Check, if response received but not successful, that it is a
                # status code we are willing to retry and we have retries left
                if resp.status_code not in _RETRY_STATUS_CODES or final_retry:
                    return resp
                else:
                    failure_cause_msg = (
                        f"Error ({resp.status_code}) for"
                        f" {resp.request.method}:{resp.url}"
                    )
            except requests.ConnectionError as ex:
                # If a connection error occurs, we can retry unless
                # this is our final attempt
                if final_retry:
                    raise ex
                else:
                    # If the exception contains request info, we can use it
                    if req := ex.request:
                        failure_cause_msg = (
                            f"Connection error ({str(ex)}) for"
                            f" {req.method}:{req.url}"
                        )
                    else:
                        failure_cause_msg = f"Connection error ({str(ex)})"

            # If we reach this point we must be attempting a retry
            retry_count += 1

            # Backoff is increased using standard exponential backoff formula
            # For standard backoff factor of one this results in backoffs of
            # [1, 2, 4, 8, ...] seconds.
            backoff = backoff_factor * (2 ** (retry_count - 1))

            # Log out failure information and retry information.
            _logger.debug(
                f"{failure_cause_msg}; "
                f"will retry in {backoff} seconds (attempt {retry_count})."
            )

            time.sleep(backoff)

        # We shouldn't reach this point due to how the loop can be exited,
        # but just in case
        raise requests.ConnectionError(
            "Unable to make connection, even after multiple attempts."
        )

    return cast(_SYNC_F, _wrapped_sync_req_func)


def _get_async_wrapper(
    req_func: _ASYNC_F, max_retries: int, backoff_factor: int
) -> _ASYNC_F:
    @wraps(req_func)
    async def _wrapped_async_req_func(*args: Any, **kwargs: Any) -> httpx.Response:
        """Wraps target HTTPX request function in retry capability.

        Adds automatic retry, backoff, and logging.
        """
        # Set default timeout if one not provided
        timeout = kwargs.get("timeout", None)
        if timeout is None or timeout is USE_CLIENT_DEFAULT:
            logging.debug(
                f"No request timeout provided,"
                f" setting to default timeout ({_DEFAULT_TIMEOUT}s)"
            )
            # Want to allow arbitrary time lengths for read and write as
            # upload/download may take a while
            kwargs["timeout"] = httpx.Timeout(
                connect=_DEFAULT_TIMEOUT,
                read=None,
                write=None,
                pool=_DEFAULT_TIMEOUT,
            )

        retry_count = 0

        while retry_count <= max_retries:
            final_retry = retry_count == max_retries

            # Attempt to make wrapped request and handle if it doesn't work
            # as expected
            try:
                resp: httpx.Response = await req_func(*args, **kwargs)

                # Check, if response received but not successful, that it is a
                # status code we are willing to retry and we have retries left
                if resp.status_code not in _RETRY_STATUS_CODES or final_retry:
                    return resp
                else:
                    failure_cause_msg = (
                        f"Error ({resp.status_code}) for"
                        f" {resp.request.method}:{resp.url}"
                    )
            except httpx.ConnectError as ex:
                # If a connection error occurs, we can retry unless
                # this is our final attempt
                if final_retry:
                    raise ex
                else:
                    # If the exception contains request info, we can use it
                    try:
                        req = ex.request
                    except RuntimeError:
                        failure_cause_msg = f"Connection error ({str(ex)})"
                    else:
                        failure_cause_msg = (
                            f"Connection error ({str(ex)}) for"
                            f" {req.method}:{req.url}"
                        )

            # If we reach this point we must be attempting a retry
            retry_count += 1

            # Backoff is increased using standard exponential backoff formula
            # For standard backoff factor of one this results in backoffs of
            # [1, 2, 4, 8, ...] seconds.
            backoff = backoff_factor * (2 ** (retry_count - 1))

            # Log out failure information and retry information.
            _logger.debug(
                f"{failure_cause_msg}; "
                f"will retry in {backoff} seconds (attempt {retry_count})."
            )

            await asyncio.sleep(backoff)

        # We shouldn't reach this point due to how the loop can be exited,
        # but just in case
        raise httpx.HTTPError(
            "Unable to make connection, even after multiple attempts."
        )

    return cast(_ASYNC_F, _wrapped_async_req_func)


# Create patched versions of the `requests` methods
request = _auto_retry_request(  #: This is needed to get pdoc to pick these up
    requests.request
)
head = _auto_retry_request(  #: This is needed to get pdoc to pick these up
    requests.head
)
get = _auto_retry_request(requests.get)  #: This is needed to get pdoc to pick these up
post = _auto_retry_request(  #: This is needed to get pdoc to pick these up
    requests.post
)
put = _auto_retry_request(requests.put)  #: This is needed to get pdoc to pick these up
patch = _auto_retry_request(  #: This is needed to get pdoc to pick these up
    requests.patch
)
delete = _auto_retry_request(  #: This is needed to get pdoc to pick these up
    requests.delete
)

__pdoc__ = {
    "request": "See `requests.request()` for more details.",
    "head": "See `requests.head()` for more details.",
    "get": "See `requests.get()` for more details.",
    "post": "See `requests.post()` for more details.",
    "put": "See `requests.put()` for more details.",
    "patch": "See `requests.patch()` for more details.",
    "delete": "See `requests.delete()` for more details.",
}


# Create patched versions of the HTTPX methods
class _AsyncClient(httpx.AsyncClient):
    # We only wrap this method in _auto_retry_request as any calls to the others
    # (post, get, etc) will make use of this. Wrapping them all would result in
    # a double retry loop, but we can't _not_ wrap request as it is often used
    # directly.
    @_auto_retry_request
    async def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
        extensions: Optional[dict] = None,
    ) -> httpx.Response:
        """See httpx.AsyncClient.request() for information."""  # noqa: D402
        return await super().request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_request(
    method: str,
    url: URLTypes,
    *,
    content: Optional[RequestContent] = None,
    data: Optional[RequestData] = None,
    files: Optional[RequestFiles] = None,
    json: Optional[Any] = None,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.request() for more information."""
    async with _AsyncClient() as client:
        return await client.request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_get(
    url: URLTypes,
    *,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.get() for more information."""
    async with _AsyncClient() as client:
        return await client.get(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_options(
    url: URLTypes,
    *,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.options() for more information."""
    async with _AsyncClient() as client:
        return await client.options(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_head(
    url: URLTypes,
    *,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.head() for more information."""
    async with _AsyncClient() as client:
        return await client.head(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_post(
    url: URLTypes,
    *,
    content: Optional[RequestContent] = None,
    data: Optional[RequestData] = None,
    files: Optional[RequestFiles] = None,
    json: Optional[Any] = None,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.post() for more information."""
    async with _AsyncClient() as client:
        return await client.post(
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_put(
    url: URLTypes,
    *,
    content: Optional[RequestContent] = None,
    data: Optional[RequestData] = None,
    files: Optional[RequestFiles] = None,
    json: Optional[Any] = None,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.put() for more information."""
    async with _AsyncClient() as client:
        return await client.put(
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_patch(
    url: URLTypes,
    *,
    content: Optional[RequestContent] = None,
    data: Optional[RequestData] = None,
    files: Optional[RequestFiles] = None,
    json: Optional[Any] = None,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.patch() for more information."""
    async with _AsyncClient() as client:
        return await client.patch(
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )


async def async_delete(
    url: URLTypes,
    *,
    params: Optional[QueryParamTypes] = None,
    headers: Optional[HeaderTypes] = None,
    cookies: Optional[CookieTypes] = None,
    auth: Union[AuthTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
    timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    extensions: Optional[dict] = None,
) -> httpx.Response:
    """See httpx.delete() for more information."""
    async with _AsyncClient() as client:
        return await client.delete(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions,
        )



"""Interface classes and methods for the message service.

Provides functionality to wrap generated GRPC code and python-friendly versions
of GRPC classes and methods.
"""
from __future__ import annotations

import asyncio
from asyncio import AbstractEventLoop
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from functools import cached_property
import logging
from math import ceil
import os
import platform
import re
from re import Pattern
import tempfile
import threading
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Dict,
    Final,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)
from urllib.parse import urlparse, urlunparse

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from grpc import RpcError, StatusCode
import msgpack
import numpy as np
import pandas as pd
import psutil
import pyarrow as pa
from pyarrow import feather
from requests.exceptions import RequestException

from bitfount.config import get_gpu_metadata
from bitfount.federated.encryption import _AESEncryption, _RSAEncryption
from bitfount.federated.exceptions import PodConnectFailedError
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.transport.config import MessageServiceConfig
from bitfount.federated.transport.exceptions import BitfountMessageServiceError
from bitfount.federated.transport.protos.messages_pb2 import (
    BitfountMessage as GrpcBitfountMessage,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    CommunicationDetails as GrpcCommunicationDetails,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    LargeStorageRequest,
    PodData,
    SuccessResponse,
)
from bitfount.federated.transport.protos.messages_pb2 import Acknowledgement
from bitfount.federated.transport.protos.messages_pb2 import BitfountTask, BitfountTasks
from bitfount.federated.transport.protos.messages_pb2_grpc import MessageServiceStub
from bitfount.federated.transport.types import CommunicationDetails
from bitfount.federated.transport.utils import _auto_retry_grpc
from bitfount.storage import (
    _async_download_data_from_s3,
    _async_upload_data_to_s3,
    _get_packed_data_object_size,
)
from bitfount.types import _S3PresignedPOSTFields, _S3PresignedPOSTURL, _S3PresignedURL
from bitfount.utils import _get_mb_from_bytes

if TYPE_CHECKING:
    from bitfount.hub.authentication_flow import BitfountSession

logger = _get_federated_logger(__name__)

# This is used for sanity checking/detecting when the message body is an S3 URL.
# However, it doesn't give us any security guarantees as this subdomain may not be
# unique to us.
_S3_NETLOC_REGEX: Final[Pattern[str]] = re.compile(
    r"-message-service-external[0-9]*\.s3\.eu-west-2\.amazonaws.com"
)


def _is_s3_netloc(netloc: str) -> bool:
    """Checks if the given netloc URL segment matches expected S3 URL structure."""
    if _S3_NETLOC_REGEX.search(netloc):
        return True
    else:
        return False


# This limit is lower than the SQS message limit intentionally.
# When messages are packaged on the message service we attach some metadata,
# the metadata will make them larger, so the python code needs a lower limit.
# So this limit ensures that our message service will accept the raw messages.
_SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES: int = 100 * 1024  # 100kb

# Maximum storage request size in message service: 3gb. This should match the value
# in bitfount/message-service:src/messaging/queue.service.ts
_MAX_STORAGE_SIZE_BYTES: Final[int] = 3 * 1024 * 1024 * 1024
_MAX_STORAGE_SIZE_MEGABYTES: Final[int] = _get_mb_from_bytes(
    _MAX_STORAGE_SIZE_BYTES
).whole

# Create fibonacci backoffs to avoid constant querying for messages. Will poll
# (60/1=60) times the first minute after the last message (i.e. 1s intervals),
# (60/2=30) times (i.e. 2s intervals) for the second minute, etc, up to a max of
# 8s polling intervals.
_POLLING_BACKOFFS: Final[Tuple[int, ...]] = tuple(
    i for i in (1, 2, 3, 5, 8) for _ in range(ceil(60 / i))
)


def _current_time() -> str:
    """Gets current time as factory for dataclass timestamps."""
    return datetime.now().isoformat()


def msgpackext_encode(obj: Any) -> Any:
    """Serialises extension objects for DataFrames or np.ndarray using pyarrow.

    Args:
        obj: An object to be serialized
    Returns:
        A msgpack compatible extension form of the object.
    """
    if isinstance(obj, np.ndarray):
        tensor = pa.Tensor.from_numpy(obj)
        sink = pa.BufferOutputStream()
        pa.ipc.write_tensor(tensor, sink)
        buf = sink.getvalue().to_pybytes()
        return msgpack.ExtType(1, buf)

    if isinstance(obj, pd.DataFrame):
        sink = pa.BufferOutputStream()
        feather.write_feather(obj, sink)
        buf = sink.getvalue().to_pybytes()
        return msgpack.ExtType(2, buf)
    return obj


def msgpackext_decode(code: int, obj: Any) -> Any:
    """Extension to deserialise DataFrame and numpy using arrow.

    Args:
        code: An integer specifying whether the object is
            a numpy array (code=1) or Dataframe (code=2).
        obj: An encoded object, likely a dictionary.

    Returns:
        The decoded form of the object.
    """
    if code == 1:
        buf = pa.py_buffer(obj)
        tensor = pa.ipc.read_tensor(buf)
        return tensor.to_numpy()
    elif code == 2:
        return feather.read_feather(pa.py_buffer(obj))
    return obj


class _BitfountMessageType(Enum):
    """Used for providing information on the expected shape of a Message.

    These can be used to quickly identify the expected type of an unstructured message
    body from the message service.

    These message types are defined in 4 locations and any changes need to be
    reflected in all locations:

    - the message service (whole proto file)
    - the python repo (whole proto file)
    - the _BitfountMessageType class (python repo)
    - the Hub/AM for audit logs
    """

    SAML_REQUEST = GrpcBitfountMessage.SAML_REQUEST
    SAML_RESPONSE = GrpcBitfountMessage.SAML_RESPONSE
    ALGORITHM_EXCHANGE = GrpcBitfountMessage.ALGORITHM_EXCHANGE
    # TODO: [BIT-1049] Should these be combined into JOB_RESPONSE?
    #       The message body itself indicates whether it was an accept or reject.
    JOB_ACCEPT = GrpcBitfountMessage.JOB_ACCEPT
    JOB_REJECT = GrpcBitfountMessage.JOB_REJECT
    JOB_REQUEST = GrpcBitfountMessage.JOB_REQUEST
    EVALUATION_RESULTS = GrpcBitfountMessage.EVALUATION_RESULTS
    KEY_EXCHANGE = GrpcBitfountMessage.KEY_EXCHANGE
    LOG_MESSAGE = GrpcBitfountMessage.LOG_MESSAGE
    # weight updates sent modeller->worker
    MODEL_PARAMETERS = GrpcBitfountMessage.MODEL_PARAMETERS
    SECURE_SHARE = GrpcBitfountMessage.SECURE_SHARE
    TRAINING_COMPLETE = GrpcBitfountMessage.TRAINING_COMPLETE
    # task complete message modeller->worker
    TASK_COMPLETE = GrpcBitfountMessage.TASK_COMPLETE
    # task abort message worker->modeller
    TASK_ABORT = GrpcBitfountMessage.TASK_ABORT
    # weight updates send worker->modeller
    TRAINING_UPDATE = GrpcBitfountMessage.TRAINING_UPDATE
    # validation metrics sent worker->modeller
    TRAINING_METRICS = GrpcBitfountMessage.TRAINING_METRICS
    # Default value for Proto
    UNDEFINED = GrpcBitfountMessage.UNDEFINED
    # OIDC identity verification related types
    OIDC_CHALLENGE = GrpcBitfountMessage.OIDC_CHALLENGE
    # Authorisation Flow Code with PKCE Response
    OIDC_AFC_PKCE_RESPONSE = GrpcBitfountMessage.OIDC_AFC_PKCE_RESPONSE
    # Device Code Response
    OIDC_DEVICE_CODE_RESPONSE = GrpcBitfountMessage.OIDC_DEVICE_CODE_RESPONSE
    # Online check from worker->modeller
    ONLINE_CHECK = GrpcBitfountMessage.ONLINE_CHECK
    # Online response from modeller->worker
    ONLINE_RESPONSE = GrpcBitfountMessage.ONLINE_RESPONSE
    # PSI Dataset exchanges between modeller and worker
    PSI_DATASET = GrpcBitfountMessage.PSI_DATASET
    # task start message modeller->worker
    TASK_START = GrpcBitfountMessage.TASK_START


class _LargeObjectRequestHandler:
    """Slim wrapper for handling requests to blob storage."""

    @staticmethod
    async def upload_large_object(
        upload_url: _S3PresignedPOSTURL,
        upload_fields: _S3PresignedPOSTFields,
        large_object: bytes,
    ) -> None:
        """Uploads the object to the upload URL.

        Args:
            upload_url: The URL to upload to.
            upload_fields: Additional data fields needed to upload the object.
            large_object: The object to upload as bytes.

        Raises:
            HTTPError: If the upload fails.
        """
        try:
            await _async_upload_data_to_s3(upload_url, upload_fields, large_object)
        except RequestException as ex:
            raise RequestException(
                f"Failed to upload message to large message storage. Cause: {str(ex)}."
            ) from ex

    @staticmethod
    async def get_large_object_from_url(url: _S3PresignedURL) -> bytes:
        """Retrieves a larger message from the provided download URL.

        Args:
            url: The URL provided on the message service.

        Returns:
            The message that was uploaded.

        Raises:
            HTTPError: If the download fails
        """
        try:
            # We can guarantee the return type is bytes as `upload_large_object`
            # takes in bytes.
            large_object: bytes = await _async_download_data_from_s3(url)
            return large_object
        except RequestException as ex:
            raise RequestException(
                f"Failed to retrieve message from large message storage."
                f" Cause: {str(ex)}."
            ) from ex


@dataclass
class _BitfountMessage:
    """A common message structure for all messages from the message service.

    This is a pythonic wrapper on the generated gRPC Proto,
    allowing us to pre-emptively build in support for upcoming features
    as well as backwards compatibility where needed.
    """

    message_type: _BitfountMessageType
    body: bytes
    recipient: str
    recipient_mailbox_id: str
    sender: str
    sender_mailbox_id: str
    # This is provided to pods for secure aggregation
    pod_mailbox_ids: Dict[str, str] = field(default_factory=dict)
    # default_factory is used here to ensure current time is set for each instantiation
    timestamp: str = field(default_factory=_current_time)
    # Provided by the message service
    receipt_handle: Optional[str] = None
    # task_id is optional for older message formats but should be used for all
    # new ones when available.
    task_id: Optional[str] = None

    @staticmethod
    async def from_rpc(message: GrpcBitfountMessage) -> _BitfountMessage:
        """Converts Proto messages into a BitfountMessage.

        Currently requires some additional logic due to the differing protos,
        FromModeller & FromPod, but hopefully those will soon be one proto.

        Args:
            message: A message from a pod or modeller.

        Returns: A BitfountMessage containing the data from the original message

        """
        message_body: bytes = message.body
        try:
            # We attempt to unpack this, as we need to know if it is:
            #   bytes  - Encrypted
            #   string - A URL to download the message body from
            unpacked_body_to_inspect: Union[str, bytes] = msgpack.loads(
                message.body, ext_hook=msgpackext_decode
            )

            # Checks if message body is an S3 file reference or local file reference
            if isinstance(unpacked_body_to_inspect, str):
                parsed_url = urlparse(unpacked_body_to_inspect)

                # If it is an S3 file reference, retrieve the actual message from S3
                if parsed_url.scheme == "https" and _is_s3_netloc(parsed_url.netloc):
                    logger.debug("Large message, retrieving from large object storage")
                    storage_url = cast(_S3PresignedURL, unpacked_body_to_inspect)
                    message_body = (
                        await _LargeObjectRequestHandler.get_large_object_from_url(
                            storage_url
                        )
                    )

                # If it is a local file reference, retrieve the actual message and
                # re-parse the message with the file contents. For windows, the letter
                # corresponding to the local drive gets assigned to the scheme, so
                # we need an additional check.
                elif (
                    (
                        not parsed_url.scheme
                        or (
                            len(parsed_url.scheme) == 1
                            and platform.system() == "Windows"
                        )
                    )
                    and not parsed_url.netloc
                    and parsed_url.path
                ):
                    logger.debug("Local message, retrieving from local storage")
                    with open(unpacked_body_to_inspect, "rb") as f:
                        message_body = f.read()
                    # deletes the tempfile after contents have been read
                    os.remove(unpacked_body_to_inspect)

                # If the parsed_url contains a netloc, but it's not a message service
                # S3 one, then raise an exception
                elif parsed_url.netloc:
                    domain_url = urlunparse(
                        (parsed_url.scheme, parsed_url.netloc, "", "", "", "")
                    )
                    raise ValueError(
                        f"Unexpected URL domain or scheme in retrieved message"
                        f" ({message.messageType}) body: {domain_url}"
                    )

                # Otherwise log that this is an unexpected message
                else:
                    logger.warning(
                        f"Received message ({message.messageType}) from"
                        f" {message.sender} with unencrypted string body"
                        f" that was not a URL."
                    )
        except (msgpack.exceptions.UnpackException, ValueError):
            logger.debug(
                "Couldn't unpack message - "
                "The message is most likely an encrypted message "
                "which is not on blob storage"
            )

        return _BitfountMessage(
            message_type=_BitfountMessageType(message.messageType),
            body=message_body,
            recipient=message.recipient,
            recipient_mailbox_id=message.recipientMailboxId,
            sender=message.sender,
            sender_mailbox_id=message.senderMailboxId,
            pod_mailbox_ids=dict(message.podMailboxIds),
            receipt_handle=message.receiptHandle,
            timestamp=message.timestamp,
            # Older messages may not have task_id field (so will either not be
            # present or default to "")
            task_id=(
                message.taskId
                if hasattr(message, "taskId") and message.taskId
                else None
            ),
        )

    def decrypt(self, aes_key: bytes) -> _DecryptedBitfountMessage:
        """Decrypt message body using AES.

        Args:
            aes_key: AES Key to use for decryption
        """
        decrypted_message_body = msgpack.loads(
            _MessageEncryption.decrypt_incoming_message(self.body, aes_key),
            ext_hook=msgpackext_decode,
        )
        return _DecryptedBitfountMessage(
            message_type=self.message_type,
            body=decrypted_message_body,
            recipient=self.recipient,
            recipient_mailbox_id=self.recipient_mailbox_id,
            sender=self.sender,
            sender_mailbox_id=self.sender_mailbox_id,
            pod_mailbox_ids=self.pod_mailbox_ids,
            timestamp=self.timestamp,
            task_id=self.task_id,
        )

    def decrypt_rsa(self, private_key: RSAPrivateKey) -> _DecryptedBitfountMessage:
        """Decrypt message body using RSA.

        Args:
            private_key: RSA private key to use for decryption.
        """
        decrypted_message_body: Any = msgpack.loads(
            _RSAEncryption.decrypt(self.body, private_key)
        )
        return _DecryptedBitfountMessage(
            message_type=self.message_type,
            body=decrypted_message_body,
            recipient=self.recipient,
            recipient_mailbox_id=self.recipient_mailbox_id,
            sender=self.sender,
            sender_mailbox_id=self.sender_mailbox_id,
            pod_mailbox_ids=self.pod_mailbox_ids,
            timestamp=self.timestamp,
            task_id=self.task_id,
        )


@dataclass
class _DecryptedBitfountMessage:
    """BitfountMessage where the body has been decrypted.

    This gives us more confidence that we have a decrypted body,
    and allows us to pass around the decrypted message body and
    metadata together easily.
    """

    message_type: _BitfountMessageType
    body: Any
    recipient: str
    recipient_mailbox_id: str
    sender: str
    sender_mailbox_id: str
    # This is provided to pods for secure aggregation
    pod_mailbox_ids: Dict[str, str] = field(default_factory=dict)
    # default_factory is used here to ensure current time is set for each instantiation
    timestamp: str = field(default_factory=_current_time)
    # task_id is optional for older message formats but should be used for all
    # new ones when available.
    task_id: Optional[str] = None


class _AutoRetryMessageServiceStub:
    """Wrapper class that adds autoretry to MessageServiceStub."""

    def __init__(self, orig: MessageServiceStub) -> None:
        self._orig = orig

        self.PodConnect = _auto_retry_grpc(orig.PodConnect)
        self.SetupTaskMailboxes = _auto_retry_grpc(orig.SetupTaskMailboxes)
        self.SendBitfountMessage = _auto_retry_grpc(orig.SendBitfountMessage)
        self.GetLargeObjectStorage = _auto_retry_grpc(orig.GetLargeObjectStorage)

        # NOTE: Auto retry for GetBitfountMessage and AcknowledgeMessage are
        #       handled separately to ensure that it is an atomic operation.
        #
        #       See _atomic_get_and_ack().
        self.GetBitfountMessage = orig.GetBitfountMessage
        self.AcknowledgeMessage = orig.AcknowledgeMessage


class _LocalGrpc(threading.local):
    """A thread-local container class for storing the gRPC stub.

    This ensures that the gRPC stub is unique per-thread as the async gRPC API is
    not thread-safe and wrapping it in synchronisation primitives would mean that
    it blocks gRPC interactions and potentially event loops.

    Because each thread will receive its own instance of this class, we want to
    set the stub to None in __init__ so that the stub instance isn't shared between
    threads.
    """

    stub: Optional[_AutoRetryMessageServiceStub]

    def __init__(self) -> None:
        self.stub = None


class _MessageService:
    """Slim wrapper around the GRPC Functions provided by the message service.

    It stops the GRPC objects which aren't very pleasant to use from being everywhere
    as well as making it easier to replace these in future if needed.

    Args:
        session: Used for authentication.
        config: The configuration for the gRPC message service.
    """

    def __init__(self, session: BitfountSession, config: MessageServiceConfig):
        self._session = session
        self._config = config
        self._use_local_storage: bool = config.use_local_storage

        self._grpc = _LocalGrpc()

    # @cached_property can't be used because it returns the _exact_ same Awaitable
    # which cannot be awaited more than once.
    @property
    async def stub(self) -> _AutoRetryMessageServiceStub:
        """The gRPC stub generated for this instance."""
        if self._grpc.stub is None:
            self._grpc.stub = _AutoRetryMessageServiceStub(await self._config.stub)

            def _custom_exception_handler(
                loop: AbstractEventLoop, context: Dict[str, Any]
            ) -> None:
                """Custom exception handler to suppress superfluous gRPC exceptions.

                See https://github.com/grpc/grpc/issues/25364.
                """
                # TODO: [BIT-2236] Remove workaround when gRPC issue resolved
                # If it is the problematic "error", ignore
                if (
                    "Exception in callback PollerCompletionQueue._handle_events"
                    in context["message"]
                ):
                    pass
                # Otherwise, pass to the default exception handler to handle
                else:
                    return loop.default_exception_handler(context)

            loop_ = asyncio.get_running_loop()
            loop_.set_exception_handler(_custom_exception_handler)
        return self._grpc.stub

    @property
    def username(self) -> str:
        """Get the authenticated username."""
        return self._session.username

    @cached_property
    def metadata(self) -> List[Tuple[str, str]]:
        """Returns metadata required for sending a BitfountMessage."""
        if api_keys := self._session.api_keys:
            return [
                # Only the first portion of the API key is used
                ("x-api-key-id", api_keys.access_key_id.split(":")[0]),
                ("x-api-key", api_keys.access_key.split(":")[0]),
            ]

        return [("token", cast(str, self._session.access_token))]

    async def connect_pod(self, pod_name: str) -> str:
        """Make the message service aware of the Pod.

        This ensures that the correct queue is set up for this pod,
        in future it may contain more metadata.

        Args:
            pod_name: The name of the pod (the part after 'username/')

        Returns:
            The created mailbox ID for this pod.

        Raises:
            BitfountMessageServiceError: If an error occurs when communicating
                                         with the message service.
        """
        processor = platform.processor()
        pod_os = platform.system()
        logical_cpu_count = psutil.cpu_count()
        total_physical_memory_bytes = psutil.virtual_memory().total
        gpu_name, available_gpus = get_gpu_metadata()

        try:
            logger.debug("Accessing gRPC stub.")
            stub = await self.stub
            logger.debug("Calling PodConnect on message service.")
            response = await stub.PodConnect(
                PodData(
                    podName=pod_name,
                    processor=processor,
                    podOS=pod_os,
                    cpuCount=logical_cpu_count,
                    totalMemoryBytes=total_physical_memory_bytes,
                    gpuCount=available_gpus,
                    gpuName=gpu_name,
                ),
                metadata=self.metadata,
            )
            logger.debug(f"PodConnect completed, response: {response}")
        except (RpcError, RequestException) as e:
            logger.debug(e, exc_info=True)
            raise BitfountMessageServiceError(
                "Unable to connect pod to message service."
                " See debug logs for more details."
            )

        # If we've not received a successful response (but no other RPC error
        # happened) then raise it here.
        if not isinstance(response, SuccessResponse):
            raise PodConnectFailedError("Unable to connect pod to message service.")

        # TODO: [BIT-960] Currently this is just hardcoded to return the pod_name
        #       (which is what the mailbox ID will actually be). [BIT-960] will have
        #       the PodConnect method actually return the generated mailbox ID so
        #       that if the approach changes in future it only needs to change on
        #       the message service side. At that point this should return the
        #       generated mailbox ID instead.
        return pod_name

    async def setup_communication_with_pods(
        self, tasks_per_pod: Dict[str, bytes]
    ) -> CommunicationDetails:
        """Called by the modeller to set up communication channels with specified pods.

        Args:
            tasks_per_pod: Mapping of pod identifiers to AES Encrypted messages.

        Returns:
            A tuple of mailbox IDs used for communications:
                - mailbox_id (str): The modeller's mailbox ID
                - worker_mailbox_ids (dict): Mapping from Pod ID to Worker Mailbox ID

        Raises:
            BitfountMessageServiceError: If an error occurs when communicating
                                         with the message service.
        """
        logger.debug(f"Sending task request to: '{tasks_per_pod.keys()}'")
        try:
            updated_tasks_per_pod = (
                await self._maybe_upload_task_to_large_object_storage(
                    tasks_per_pod.copy()
                )
            )
            # Build the tasks
            tasks = []
            for pod_identifier, aes_encrypted_task in tasks_per_pod.items():
                if updated_tasks_per_pod[pod_identifier] == aes_encrypted_task:
                    tasks.append(
                        BitfountTask(
                            podIdentifier=pod_identifier,
                            encryptedTask=aes_encrypted_task,
                        )
                    )
                else:
                    tasks.append(
                        BitfountTask(
                            podIdentifier=pod_identifier,
                            taskURL=updated_tasks_per_pod[pod_identifier],
                        )
                    )

            communication_details: GrpcCommunicationDetails = await (
                await self.stub
            ).SetupTaskMailboxes(
                BitfountTasks(tasks=tasks),
                metadata=self.metadata,
            )
        except (RpcError, RequestException) as e:
            logger.debug(e, exc_info=True)
            raise BitfountMessageServiceError(
                "Unable to setup communication with target pods."
                " See debug logs for more details."
            )

        # communication_details.podMailboxIds is actually the _worker_ mailbox_ids
        return CommunicationDetails(
            communication_details.mailboxId,
            dict(communication_details.podMailboxIds),
            communication_details.taskId,
        )

    async def _maybe_upload_task_to_large_object_storage(
        self, tasks_per_pod: Dict[str, bytes]
    ) -> Dict[str, bytes]:
        """Uploads the task requests to large object storage if it's too big."""
        for pod_identifier, aes_encrypted_task in tasks_per_pod.items():
            # Larger messages are sent as references to large object storage
            if len(aes_encrypted_task) > _SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES:
                # Check that message body isn't above the maximum storage size
                task_size = _get_packed_data_object_size(aes_encrypted_task)
                if task_size > _MAX_STORAGE_SIZE_BYTES:
                    raise ValueError(
                        f"Message body is too large to upload: "
                        f"expected max {_MAX_STORAGE_SIZE_MEGABYTES} megabytes, "
                        f"got {_get_mb_from_bytes(task_size).fractional} megabytes."
                    )
                if "/" in pod_identifier:
                    pod_name = pod_identifier.split("/")[1]
                else:
                    pod_name = pod_identifier
                (
                    upload_url,
                    upload_fields,
                ), download_url = await self._get_large_object_storage_details(
                    self._session._client_id, task_size, None, pod_name
                )
                await _LargeObjectRequestHandler.upload_large_object(
                    upload_url,
                    upload_fields,
                    aes_encrypted_task,
                )
                # # Currently necessary, as the message service is expecting bytes
                tasks_per_pod[pod_identifier] = msgpack.dumps(
                    download_url, default=msgpackext_encode
                )
        # Small messages can just go on the message service directly
        # so we just return them.
        return tasks_per_pod

    async def poll_for_messages(
        self,
        mailbox_id: str,
        stop_event: threading.Event,
    ) -> AsyncGenerator[_BitfountMessage, None]:
        """Polls target mailbox for message.

        Will keep retrying up to `self.minutes_before_timeout_no_messages`
        or until a message is received, whichever is sooner.

        Will attempt to workaround transient connection exceptions by retrying but
        if exceptions are repeatedly raised then it will throw the exception.

        Args:
            mailbox_id: The mailbox to poll for messages.
            stop_event: A threading Event that indicates to stop polling.

        Yields:
            BitfountMessage: the next message received
        """
        counter = 0

        # Create stop_event future to use for all retrievals
        stop_event_wait = asyncio.get_running_loop().run_in_executor(
            None, stop_event.wait
        )

        while True:
            # Check stop event before trying to retrieve messages
            if stop_event.is_set():
                logger.debug("Stopping polling for messages.")
                break

            # Wait for either a message or for the stop event to be set
            get_message_task = asyncio.create_task(
                self._get_message(mailbox_id), name="get_message"
            )
            awaitables: List[Awaitable] = [get_message_task, stop_event_wait]
            done, pending = await asyncio.wait(
                awaitables, return_when=asyncio.FIRST_COMPLETED
            )

            if get_message_task in done:
                message = get_message_task.result()

                if message:
                    # Reset counter
                    counter = 0
                    # Yield as soon as we've retrieved a message
                    yield message
                else:
                    counter += 1
                    # Cap counter at the maximum `backoffs` entry
                    counter = min(counter, len(_POLLING_BACKOFFS) - 1)

                # Check stop event again before trying to sleep
                if stop_event.is_set():
                    logger.debug("Stopping polling for messages.")
                    break

                # Sleep for backoff time until trying again
                await asyncio.sleep(_POLLING_BACKOFFS[counter])
            else:
                # The stop event has been triggered
                logger.debug("Stopping polling for messages.")
                get_message_task.cancel()
                break

    async def _get_message(self, mailbox_id: str) -> Optional[_BitfountMessage]:
        """Retrieve message from mailbox if available.

        If none are available then it returns None.

        Other errors are raised if they occur repeatedly.

        Args:
            mailbox_id: Mailbox to check for messages.

        Returns:
            An BitfountMessage if one is available, else None.

        Raises:
            BitfountMessageServiceError: If an error occurs when multiple times
                                         when communicating with the message service.
        """
        # Try to retrieve a message and acknowledge message in one atomic step if
        # there is one available

        try:
            message: GrpcBitfountMessage = await self._atomic_get_and_ack(mailbox_id)
        except (RpcError, RequestException) as ex:
            if isinstance(ex, RpcError) and ex.code() == StatusCode.NOT_FOUND:
                # This indicates that no message was waiting and is an expected "error"
                logging.debug(f"No message available on {mailbox_id}")
                return None
            else:
                logging.debug(ex, exc_info=True)
                logging.error(
                    f"Exceeded maximum attempts when retrieving message "
                    f"from {mailbox_id}"
                )
                raise BitfountMessageServiceError(
                    "Issue retrieving message from mailbox."
                    " See debug log for details."
                )

        b_message = await _BitfountMessage.from_rpc(message)
        logger.debug(
            f"Retrieved {b_message.message_type} message"
            f" from {b_message.sender}"
            f" on mailbox {mailbox_id}"
        )
        return b_message

    # We also explicitly add NOT_FOUND to the GetBitfountMessage non-retries
    # as this is used to denote "no message found"
    @_auto_retry_grpc(additional_no_retry_status_codes=[StatusCode.NOT_FOUND])
    async def _atomic_get_and_ack(
        self, mailbox_id: str, timeout: Optional[float] = None
    ) -> GrpcBitfountMessage:
        """Get and acknowledge gRPC message in one atomic operation.

        Will either return the retrieved and ACKed message or raise an RpcError
        if something goes wrong at either stage or if there is no message available.
        """
        logger.debug(f"Attempting message retrieval from {mailbox_id}")
        message = await (await self.stub).GetBitfountMessage(
            # Only mailboxId is needed here because that's all that's used on the
            # message service side. Despite the other args not being "optional"
            # we can let the default values be used.
            GrpcCommunicationDetails(mailboxId=mailbox_id),
            metadata=self.metadata,
            timeout=timeout,
        )
        logger.debug(f"Message retrieved (id={id(message)})")

        delete_mailbox = False
        if message.messageType == _BitfountMessageType.TASK_COMPLETE.value:
            logging.debug(
                "Received a TASK_COMPLETE message - "
                "asking message service to tidy up."
            )
            delete_mailbox = True

        logger.debug(f"Attempting to ACK message (id={id(message)})")
        await (await self.stub).AcknowledgeMessage(
            Acknowledgement(
                mailboxId=mailbox_id,
                receiptHandle=message.receiptHandle,
                deleteMailbox=delete_mailbox,
            ),
            metadata=self.metadata,
            timeout=timeout,
        )
        logger.debug(f"ACKed message (id={id(message)})")

        return message

    async def send_message(
        self,
        message: _BitfountMessage,
        already_packed: bool = False,
    ) -> SuccessResponse:
        """Send a message.

        Args:
            message (BitfountMessage): The message to send
            already_packed (bool): Whether or not the message is already in required
                bytes format

        Returns:
            SuccessResponse if the message was successfully sent

        Raises:
            BitfountMessageServiceError: If an error occurs when communicating
                                         with the message service.
        """
        if not already_packed:
            message.body = msgpack.dumps(
                message.body,
                default=msgpackext_encode,
            )

        if self._use_local_storage:
            message.body = self._save_object_to_local_storage(message)
        else:
            message.body = await self._maybe_upload_to_large_object_storage(message)

        try:
            logger.debug(
                f"Sending {message.message_type}"
                f" to {message.recipient} (id={id(message)})"
            )
            send_resp = await (await self.stub).SendBitfountMessage(
                GrpcBitfountMessage(
                    messageType=message.message_type.value,
                    body=message.body,
                    recipient=message.recipient,
                    recipientMailboxId=message.recipient_mailbox_id,
                    sender=message.sender,
                    senderMailboxId=message.sender_mailbox_id,
                    timestamp=message.timestamp,
                    podMailboxIds=message.pod_mailbox_ids,
                    taskId=message.task_id,
                ),
                metadata=self.metadata,
            )
            logger.debug(
                f"Sent {message.message_type} to {message.recipient}"
                f" (id={id(message)}): {send_resp}"
            )
            return send_resp
        except (RpcError, RequestException) as e:
            logger.debug(
                f"Error sending {message.message_type}"
                f" to {message.recipient} (id={id(message)})"
            )
            logger.debug(e, exc_info=True)
            raise BitfountMessageServiceError(
                "Encountered problem when sending message."
                " See debug logs for more details."
            )

    async def _maybe_upload_to_large_object_storage(
        self, message: _BitfountMessage
    ) -> bytes:
        """Uploads the message body to large object storage if it's too big.

        Args:
            message: the message containing the body to upload

        Returns:
            Either the original message body, or a URL that can be used to download it
        """
        # Larger messages are sent as references to large object storage
        if len(message.body) > _SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES:
            # Check that message body isn't above the maximum storage size
            message_body_size = _get_packed_data_object_size(message.body)
            if message_body_size > _MAX_STORAGE_SIZE_BYTES:
                raise ValueError(
                    f"Message body is too large to upload: "
                    f"expected max {_MAX_STORAGE_SIZE_MEGABYTES} megabytes, "
                    f"got {_get_mb_from_bytes(message_body_size).fractional} megabytes."
                )

            pod_name = None
            if message.sender and "/" in message.sender:
                pod_name = message.sender.split("/")[1]
            (
                upload_url,
                upload_fields,
            ), download_url = await self._get_large_object_storage_details(
                message.sender_mailbox_id, message_body_size, message.task_id, pod_name
            )
            logger.debug(
                f"Uploading contents of {message.message_type} to object storage"
            )
            await _LargeObjectRequestHandler.upload_large_object(
                upload_url, upload_fields, message.body
            )

            # Currently necessary, as the message service is expecting bytes
            return msgpack.dumps(download_url, default=msgpackext_encode)
        # Small messages can just go on the message service directly
        return message.body

    async def _get_large_object_storage_details(
        self,
        mailbox_id: str,
        message_size: int,
        task_id: Union[str, None],
        pod_name: Optional[str] = None,
    ) -> Tuple[Tuple[_S3PresignedPOSTURL, _S3PresignedPOSTFields], _S3PresignedURL]:
        """Get an upload and download URL for larger message bodies.

        Args:
            mailbox_id: Your mailbox id for the current task.
            message_size: The size of the message to upload (an upper limit).
            task_id: The ID associated with the task. Must be supplied but can be None.
            pod_name: Name of the pod that you're acting as, if relevant.

        Returns:
            Tuple of upload handler, download URL.

        Raises:
            BitfountMessageServiceError: If an error occurs when communicating
                with the message service.
        """
        try:
            blob_storage_data = await (await self.stub).GetLargeObjectStorage(
                LargeStorageRequest(
                    senderMailboxId=mailbox_id,
                    podName=pod_name,
                    contentSize=message_size,
                    taskId=task_id,
                ),
                metadata=self.metadata,
            )
            return (
                (
                    cast(_S3PresignedPOSTURL, blob_storage_data.uploadUrl),
                    cast(_S3PresignedPOSTFields, dict(blob_storage_data.uploadFields)),
                ),
                cast(_S3PresignedURL, blob_storage_data.downloadUrl),
            )
        except (RpcError, RequestException) as e:
            logger.debug(e, exc_info=True)
            raise BitfountMessageServiceError(
                "Unable to acquire large object storage."
                " See debug logs for more details."
            )

    @staticmethod
    def _save_object_to_local_storage(message: _BitfountMessage) -> bytes:
        """Saves object to local storage and returns filename.

        Args:
            message: the message containing the body to upload

        Returns:
            (bytes): A path to the file where the local object has been stored.
        """
        # we use mkstemp here to ensure we have complete control over when the file gets
        # deleted on every OS - unlike e.g. NamedTemporaryFile
        handle, local_tempfile = tempfile.mkstemp()
        # write to the open file handler rather than finding the file by name
        with os.fdopen(handle, "wb") as f:
            f.write(message.body)
        return msgpack.dumps(local_tempfile, default=msgpackext_encode)


class _MessageEncryption:
    """Shared functions for handling encryption/decryption of messages.

    These handle encrypting or decrypting a single or multiple messages
    at the same time using different keys (for different pods)
    """

    @staticmethod
    def encrypt_outgoing_message(body: bytes, encryption_key: bytes) -> bytes:
        """Encrypt `body` using `encryption_key`."""
        body, nonce = _AESEncryption.encrypt(encryption_key, body)
        body += nonce
        return body

    @staticmethod
    def decrypt_incoming_message(
        body: bytes,
        encryption_key: bytes,
    ) -> bytes:
        """Decrypt incoming message."""
        body, nonce = body[:-12], body[-12:]
        body = _AESEncryption.decrypt(encryption_key, nonce, body)

        return body



"""Classes concerning data structures.

DataStructures provide information about the columns of a BaseSource for a specific
Modelling Job.
"""
from __future__ import annotations

from dataclasses import dataclass
import inspect
import logging
from os import PathLike
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Type,
    Union,
    cast,
)

import desert
from marshmallow import Schema as MarshmallowSchema
from marshmallow import fields, post_load
import yaml

from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasplitters import DatasetSplitter
from bitfount.data.exceptions import DataStructureError
from bitfount.data.schema import BitfountSchema, TableSchema
from bitfount.data.types import (
    CategoricalRecord,
    ContinuousRecord,
    DataSplit,
    ImageRecord,
    SchemaOverrideMapping,
    SemanticType,
    StrDictField,
    TextRecord,
    _SemanticTypeRecord,
    _SemanticTypeValue,
)
from bitfount.transformations.base_transformation import TRANSFORMATION_REGISTRY
from bitfount.transformations.batch_operations import BatchTimeOperation
from bitfount.transformations.parser import TransformationsParser
from bitfount.types import (
    T_FIELDS_DICT,
    T_NESTED_FIELDS,
    _BaseSerializableObjectMixIn,
    _JSONDict,
)
from bitfount.utils import _add_this_to_list

if TYPE_CHECKING:
    from bitfount.data.datasources.base_source import BaseSource
    from bitfount.runners.config_schemas import (
        DataStructureAssignConfig,
        DataStructureSelectConfig,
        DataStructureTableConfig,
        DataStructureTransformConfig,
    )

logger = logging.getLogger(__name__)

DEFAULT_IMAGE_TRANSFORMATIONS: List[Union[str, _JSONDict]] = [
    {"Resize": {"height": 224, "width": 224}},
    "Normalize",
    "ToTensorV2",
]

_registry: Dict[str, Type[BaseDataStructure]] = {}
registry: Mapping[str, Type[BaseDataStructure]] = MappingProxyType(_registry)


@dataclass
class BaseDataStructure:
    """Base DataStructure class."""

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
            _registry[cls.__name__] = cls


@dataclass
class DataStructure(BaseDataStructure, _BaseSerializableObjectMixIn):
    """Information about the columns of a BaseSource.

    This component provides the desired structure of data
    to be used by discriminative machine learning models.

    :::note

    If the datastructure includes image columns, batch transformations will be applied
    to them.

    :::

    Args:
        table: The table in the Pod schema to be used for local data. If executing a
            remote task, this should be a mapping of Pod names to table names.
            Defaults to None.
        query: The sql query that needs to be applied to the data.
            It should be a string if it is used for local data or a mapping
            of Pod names to the queries. Defaults to None.
        schema_types_override: A mapping that defines the new data types that
            will be returned after the sql query is executed. For a local training
            task it will be a mapping of column names to their types, for a remote
            task it will be a mapping of the Pod name to the new columns and types.
            If a column is defined as `categorical`, the mapping should include a
            mapping to the categories. Required if a sql query is provided.
            E.g. {'Pod_id': {'categorical': [{'col1': {'value_1':0, 'value_2': 1}}],
            'continuous': ['col2']} for remote training or {'categorical':
            [{'col1': {'value_1':0, 'value_2': 1}}],'continuous': ['col2']}
            for local training. Defaults to None.
        target: The training target column or list of columns.
        ignore_cols: A list of columns to ignore when getting the
            data. Defaults to None.
        selected_cols: A list of columns to select when getting the
            data. Defaults to None.
        data_splitter: Approach used for splitting the data into training, test,
            validation. Defaults to None.
        loss_weights_col: A column name which provides a weight to be given
            to each sample in loss function. Defaults to None.
        multihead_col: A categorical column whereby the number of unique values
            will determine number of heads in a Neural Network. Used
            for multitask training. Defaults to None.
        multihead_size: The number of uniques values in the `multihead_col`.
            Used for multitask training. Required if `multihead_col` is
            provided. Defaults to None.
        ignore_classes_col: A column name denoting which classes to ignore
            in a multilabel multiclass classification problem. Each value is
            expected to contain a list of numbers corresponding to the indices of
            the classes to be ignored as per the order provided in `target`.
            E.g. [0,2,3]. An empty list can be provided (e.g. []) to avoid ignoring
            any classes for some samples. Defaults to None.
        image_cols: A list of columns that will be treated as images in the data.
        batch_transforms: A dictionary of transformations to apply to batches.
            Defaults to None.
        dataset_transforms: A dictionary of transformations to apply to
            the whole dataset. Defaults to None.

    Raises:
        DataStructureError: If 'sql_query' is provided as well as either `selected_cols`
            or `ignore_cols`.
        DataStructureError: If both `ignore_cols` and `selected_cols` are provided.
        DataStructureError: If the `multihead_col` is provided without `multihead_size`.
        ValueError: If a batch transformation name is not recognised.

    """

    fields_dict = {
        "table": StrDictField(allow_none=True),
        "query": StrDictField(allow_none=True),
        "schema_types_override": fields.Dict(allow_none=True),
        "target": fields.Raw(allow_none=True),
        "ignore_cols": fields.List(fields.Str(), allow_none=True),
        "selected_cols": fields.List(fields.Str(), allow_none=True),
        "loss_weights_col": fields.Str(allow_none=True),
        "multihead_col": fields.Str(allow_none=True),
        "multihead_size": fields.Int(allow_none=True),
        "ignore_classes_col": fields.Str(allow_none=True),
        "batch_transforms": fields.List(
            fields.Dict(
                keys=fields.Str(),
                values=fields.Dict(keys=fields.Str()),
            ),
            allow_none=True,
        ),
        "dataset_transforms": fields.List(
            fields.Dict(
                keys=fields.Str(),
                values=fields.Dict(keys=fields.Str()),
            ),
            allow_none=True,
        ),
    }  # type: ClassVar[T_FIELDS_DICT]
    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
    table: Optional[Union[str, Mapping[str, str]]] = None
    query: Optional[Union[str, Mapping[str, str]]] = None
    schema_types_override: Optional[
        Union[SchemaOverrideMapping, Mapping[str, SchemaOverrideMapping]]
    ] = None
    target: Optional[Union[str, List[str]]] = None
    # Mypy errors ignored below. For more details see `config_schemas.py`
    ignore_cols: List[str] = desert.field(  # type: ignore[assignment] # Reason: above
        fields.List(fields.String()), default_factory=list
    )
    selected_cols: List[str] = desert.field(  # type: ignore[assignment] # Reason: above
        fields.List(fields.String()), default_factory=list
    )
    data_splitter: Optional[DatasetSplitter] = None
    loss_weights_col: Optional[str] = None
    multihead_col: Optional[str] = None
    multihead_size: Optional[int] = None
    ignore_classes_col: Optional[str] = None
    image_cols: Optional[List[str]] = None
    batch_transforms: Optional[List[Dict[str, _JSONDict]]] = None
    dataset_transforms: Optional[List[Dict[str, _JSONDict]]] = None

    def __post_init__(self) -> None:
        self.class_name = type(self).__name__
        if self.table and self.query:
            raise DataStructureError(
                "Invalid parameter specification. "
                "Please provide either table name (table) or "
                "query to execute (query), not both. "
            )
        if not self.table and not self.query:
            raise DataStructureError(
                "Invalid parameter specification. "
                "Please provide one of table name (table) or "
                "query to execute (query). "
            )
        if self.query and (self.selected_cols or self.ignore_cols):
            raise DataStructureError(
                "If a query is specified, the columns needed for training "
                "should be defined in the query."
            )
        if self.query and self.schema_types_override is None:
            raise DataStructureError(
                "Invalid parameter specification. "
                "Please provide the schema override mapping along with the query."
            )
        if self.schema_types_override:
            self._validate_schema_features()

        if self.selected_cols and self.ignore_cols:
            raise DataStructureError(
                "Invalid parameter specification. "
                "Please provide either columns to select (selected_cols) or "
                "to ignore (ignore_cols), not both."
            )
        if self.multihead_col and self.multihead_size is None:
            raise DataStructureError("Please provide the size of the multihead column.")
        if self.dataset_transforms is not None:
            self.set_columns_after_transformations(self.dataset_transforms)
        self._force_stype: MutableMapping[_SemanticTypeValue, List[str]] = {}
        if self.image_cols:
            self._force_stype["image"] = self.image_cols

        if self.batch_transforms is None and self.image_cols:
            default_image_transformations = []
            for col in self.image_cols:
                for step in DataSplit:
                    default_image_transformations.append(
                        {
                            "albumentations": {
                                "arg": col,
                                "output": True,
                                "transformations": DEFAULT_IMAGE_TRANSFORMATIONS,
                                "step": step.value,
                            }
                        }
                    )
            self.batch_transforms = default_image_transformations

        # Ensure specified batch transformations are all valid transformations
        if self.batch_transforms is not None:
            invalid_batch_transforms = []
            for _dict in self.batch_transforms:
                for tfm in _dict:
                    if tfm not in TRANSFORMATION_REGISTRY:
                        invalid_batch_transforms.append(tfm)
            if invalid_batch_transforms:
                raise ValueError(
                    f"The following batch transformations are not recognised: "
                    f"{', '.join(sorted(invalid_batch_transforms))}."
                )

        # Create mapping of all feature names used in training together with the
        # corresponding semantic type. This is the final mapping that will be used
        # to decide which features will be actually be used.
        self.selected_cols_w_types: Dict[_SemanticTypeValue, List[str]] = {}

    @classmethod
    def create_datastructure(
        cls,
        table_config: DataStructureTableConfig,
        select: DataStructureSelectConfig,
        transform: DataStructureTransformConfig,
        assign: DataStructureAssignConfig,
    ) -> DataStructure:
        """Creates a datastructure based on the yaml config.

        Args:
            table_config: The table in the Pod schema to be used for local data.
                If executing a remote task, this should a mapping of Pod names
                to table names.
            select: The configuration for columns to be included/excluded
                from the `DataStructure`.
            transform: The configuration for dataset and batch transformations
                to be applied to the data.
            assign: The configuration for special columns in the `DataStructure`.

        Returns:
              A `DataStructure` object.
        """
        if select.include and select.exclude:
            raise DataStructureError(
                "Please provide either columns to include or to exclude from data"
                ", not both."
            )
        ignore_cols = select.exclude if select.exclude is not None else []
        selected_cols = select.include if select.include is not None else []
        return cls(
            table=table_config.table,
            query=table_config.query,
            schema_types_override=table_config.schema_types_override,
            target=assign.target,
            ignore_cols=ignore_cols,
            selected_cols=selected_cols,
            loss_weights_col=assign.loss_weights_col,
            multihead_col=assign.multihead_col,
            ignore_classes_col=assign.ignore_classes_col,
            image_cols=assign.image_cols,
            batch_transforms=transform.batch,
            dataset_transforms=transform.dataset,
        )

    @classmethod
    def load_from_file(cls, file_path: Union[str, PathLike]) -> DataStructure:
        """Loads DataStructure from yaml file.

        Args:
            file: A yaml file with the `DataStructure` configuration.

        Returns:
            The loaded `DataStructure`.
        """
        with open(file_path, "r") as f:
            datastructure_yaml = yaml.safe_load(f)
        datastructure: DataStructure = cls._Schema().load(datastructure_yaml)
        return datastructure

    def get_table_name(self, pod_identifier: Optional[str] = None) -> str:
        """Returns the relevant table name of the `DataStructure`.

        Returns:
            The table name of the `DataStructure` corresponding to the `pod_identifier`
            provided or just the local table name if running locally.

        Raises:
            ValueError: If the `pod_identifier` is not provided and there are different
                table names for different pods.
        """
        if isinstance(self.table, str):
            return self.table
        elif isinstance(self.table, dict) and pod_identifier:
            return cast(str, self.table[pod_identifier])

        raise ValueError("No pod identifier provided for multi-pod datastructure.")

    def get_pod_identifiers(self) -> Optional[List[str]]:
        """Returns a list of pod identifiers specified in the `table` attribute.

        If there are no pod identifiers specified, returns None.
        """
        if self.table is not None:
            if isinstance(self.table, str):
                return None
            else:
                pod_identifiers = list(self.table)
        elif self.query is not None:
            if isinstance(self.query, str):
                return None
            else:
                pod_identifiers = list(self.query)
        else:
            return None
        return pod_identifiers

    def get_columns_ignored_for_training(self, table_schema: TableSchema) -> List[str]:
        """Adds all the extra columns that will not be used in model training.

        Args:
            table_schema: The schema of the table.

        Returns:
            ignore_cols_aux: A list of columns that will be ignored when
                training a model.
        """
        if self.selected_cols:
            self.ignore_cols = [
                feature
                for feature in table_schema.get_feature_names()
                if feature not in self.selected_cols
            ]
        ignore_cols_aux = self.ignore_cols[:]
        ignore_cols_aux = _add_this_to_list(self.target, ignore_cols_aux)
        ignore_cols_aux = _add_this_to_list(self.loss_weights_col, ignore_cols_aux)
        ignore_cols_aux = _add_this_to_list(self.ignore_classes_col, ignore_cols_aux)
        return ignore_cols_aux

    def set_training_input_size(self, schema: TableSchema) -> None:
        """Get the input size for model training.

        Args:
            schema: The schema of the table.
            table_name: The name of the table.
        """
        self.input_size = len(
            [
                col
                for col in schema.get_feature_names()
                if col not in self.get_columns_ignored_for_training(schema)
                and col not in schema.get_feature_names(SemanticType.TEXT)
            ]
        )

    def set_training_column_split_by_semantic_type(self, schema: TableSchema) -> None:
        """Sets the column split by type from the schema.

        This method splits the selected columns from the dataset
        based on their semantic type.

        Args:
            schema: The `TableSchema` for the data.
        """
        if not self.selected_cols and not self.ignore_cols:
            # If neither selected_cols or ignore_cols are provided,
            # select all columns from schema,
            self.selected_cols = schema.get_feature_names()
        elif self.selected_cols:
            # Make sure we set self.ignore_cols
            self.ignore_cols = [
                feature
                for feature in schema.get_feature_names()
                if feature not in self.selected_cols
            ]
        else:
            # Make sure we set self.selected_cols
            self.selected_cols = [
                feature
                for feature in schema.get_feature_names()
                if feature not in self.ignore_cols
            ]
        if self.target and self.target not in self.selected_cols:
            self.selected_cols = _add_this_to_list(self.target, self.selected_cols)
        # Get the list of all columns ignored for training
        ignore_cols_aux = self.get_columns_ignored_for_training(schema)

        # Populate mapping of all feature names used in training
        # together with the corresponding semantic type
        for stype, features in schema.features.items():
            columns_stype_list = list(cast(Dict[str, _SemanticTypeRecord], features))
            self.selected_cols_w_types[cast(_SemanticTypeValue, stype)] = [
                col for col in columns_stype_list if col not in ignore_cols_aux
            ]
        # Add mapping to empty list for all stypes not present
        # in the current datastructure
        all_stypes = [stype.value for stype in SemanticType]
        for stype in all_stypes:
            if stype not in self.selected_cols_w_types:
                self.selected_cols_w_types[cast(_SemanticTypeValue, stype)] = []

        # Get the number of images present in the datastructure.
        self.number_of_images = len(self.image_cols) if self.image_cols else 0

    def set_columns_after_transformations(
        self, transforms: List[Dict[str, _JSONDict]]
    ) -> None:
        """Updates the selected/ignored columns based on the transformations applied.

        It updates `self.selected_cols` by adding on the new names of columns after
        transformations are applied, and removing the original columns unless
        explicitly specified to keep.

        Args:
            transforms: A list of transformations to be applied to the data.
        """
        for tfm in transforms:
            for key, value in tfm.items():
                if key == "convert_to":
                    # Column name doesn't change if we only convert type.
                    pass
                else:
                    # Check to see if any original columns are marked to keep
                    original_cols_to_keep = value.get("keep_original", [])

                    # Make a list of all the columns to be discarded
                    if isinstance(value["col"], str):
                        value["col"] = [value["col"]]
                    discard_columns = [
                        col for col in value["col"] if col not in original_cols_to_keep
                    ]
                    new_columns = [f"{col}_{key}" for col in value["col"]]
                    # Error raised in the pods if we set both ignore_cols
                    # and selected_cols here.
                    if self.selected_cols:
                        self.selected_cols.extend(new_columns)
                    else:
                        self.ignore_cols.extend(discard_columns)
                    self.selected_cols = [
                        col for col in self.selected_cols if col not in discard_columns
                    ]

    def apply_dataset_transformations(self, datasource: BaseSource) -> BaseSource:
        """Applies transformations to whole dataset.

        Args:
            datasource: The `BaseSource` object to be transformed.

        Returns:
            datasource: The transformed datasource.
        """
        if self.dataset_transforms:
            # TODO: [BIT-1167] Process dataset transformations
            raise NotImplementedError()

        return datasource

    def get_batch_transformations(self) -> Optional[List[BatchTimeOperation]]:
        """Returns batch transformations to be performed as callables.

        Returns:
            A list of batch transformations to be passed to
                TransformationProcessor.
        """
        if self.batch_transforms is not None:
            parser = TransformationsParser()
            transformations, _ = parser.deserialize_transformations(
                self.batch_transforms
            )
            return cast(List[BatchTimeOperation], transformations)
        return None

    def _validate_schema_features(self) -> None:
        """Validate that the override contains encoding of categorical features."""
        self.schema_types_override = cast(
            Union[SchemaOverrideMapping, Mapping[str, SchemaOverrideMapping]],
            self.schema_types_override,
        )
        # If we validate the schema features, then it means it must be a mapping

        if isinstance(self.query, str):
            # That means that the datastructure is intended for local training.
            if "categorical" in self.schema_types_override:
                # For local training, categorical features should be
                # defined inside a dictionary.
                for item in self.schema_types_override["categorical"]:
                    if not isinstance(item, dict):
                        raise DataStructureError(
                            "Categorical features should be defined as a dictionary "
                            "with the encodings of each attribute defined. "
                            "Please specify encodings."
                        )
        elif isinstance(self.query, dict):
            for pod_id in self.query:
                if pod_id not in self.schema_types_override:
                    raise DataStructureError(
                        f"The pod id:{pod_id} given to the query was "
                        f"not found in the schema override. Please "
                        f"update the schema override types with the pod_id "
                        f"in which the schema types override will be applied. "
                    )
            for overrides in self.schema_types_override.values():
                overrides = cast(SchemaOverrideMapping, overrides)
                if "categorical" in overrides:
                    for item in overrides["categorical"]:
                        if not isinstance(item, dict):
                            raise DataStructureError(
                                "Categorical features should be defined as a "
                                "dictionary with the encodings of each attribute "
                                "defined. Please specify encodings."
                            )

    def _override_schema(
        self,
        datasource: Optional[BaseSource] = None,
        pod_identifier: Optional[str] = None,
    ) -> TableSchema:
        """Method to override the pod schema when a sql query is given."""
        self.schema_types_override = cast(
            Union[SchemaOverrideMapping, Mapping[str, SchemaOverrideMapping]],
            self.schema_types_override,
        )
        if isinstance(self.query, str):
            feature_overrides = cast(SchemaOverrideMapping, self.schema_types_override)
        elif isinstance(self.query, dict) and pod_identifier:
            feature_overrides = cast(
                SchemaOverrideMapping, self.schema_types_override[pod_identifier]  # type: ignore[index] # Reason: An error will be caught by the _validate_schema_features. # noqa: B950
            )
        else:
            raise ValueError(
                "No query or dictionary of pod_identifiers to queries was given."
            )

        table_name = "data"
        table_dtypes = None
        if datasource and datasource._data_is_loaded:
            # Data is loaded into a dataframe by the time we get here,
            # so we use the DataFrameSource to get the dtypes.
            table_dtypes = datasource.get_dtypes()

        self.selected_cols = []
        schema = TableSchema(name=table_name)
        self.table = table_name
        # At this point feature_overrides should be a mapping
        # from column names to semantic types.
        for s_type, feature_list in feature_overrides.items():
            stype = SemanticType(s_type)
            if stype == SemanticType.CONTINUOUS:
                for feature_name in feature_list:
                    # Feature_name should be string for continuous attributes
                    feature_name = cast(str, feature_name)
                    if table_dtypes:
                        dtype = table_dtypes[feature_name]
                    else:
                        # This will only be used for setting the model parameters.
                        dtype = "int"
                    self.selected_cols.append(feature_name)
                    ContinuousRecord.add_record_to_schema(
                        schema,
                        feature_name=feature_name,
                        dtype=dtype,
                    )
            elif stype == SemanticType.IMAGE:
                for feature_name in feature_list:
                    # Feature_name should be string for image attributes
                    feature_name = cast(str, feature_name)
                    if table_dtypes:
                        dtype = table_dtypes[feature_name]
                    else:
                        # This will only be used for setting the model parameters.
                        dtype = "image"
                    self.selected_cols.append(feature_name)
                    if (
                        "image" not in schema.features
                        or feature_name not in schema.features["image"]
                    ):
                        ImageRecord.add_record_to_schema(
                            schema,
                            feature_name=feature_name,
                            dtype=dtype,
                        )
            elif stype == SemanticType.TEXT:
                for feature_name in feature_list:
                    # Feature_name should be string for text attributes
                    feature_name = cast(str, feature_name)
                    if table_dtypes:
                        dtype = table_dtypes[feature_name]
                    else:
                        # This will only be used for setting the model parameters.
                        dtype = "str"
                    self.selected_cols.append(feature_name)
                    TextRecord.add_record_to_schema(
                        schema,
                        feature_name=feature_name,
                        dtype=dtype,
                    )
            elif stype == SemanticType.CATEGORICAL:
                for feature in feature_list:
                    # For categorical attributes, the `feature` here will be a dict.
                    feature_name, encodings = next(iter(cast(dict, feature).items()))
                    feature_name = cast(str, feature_name)
                    if table_dtypes:
                        dtype = table_dtypes[feature_name]
                    else:
                        # This will only be used for setting the model parameters.
                        dtype = "str"
                    self.selected_cols.append(feature_name)
                    CategoricalRecord.add_record_to_schema(
                        schema,
                        feature_name=feature_name,
                        dtype=dtype,
                    )
                    schema.features["categorical"][
                        feature_name
                    ].encoder.add_values_with_encoding(encodings)
        return schema

    def get_table_schema(
        self,
        schema: BitfountSchema,
        pod_identifier: Optional[str] = None,
        datasource: Optional[BaseSource] = None,
    ) -> TableSchema:
        """Returns the table schema based on the datastructure arguments.

        This will return either the new schema defined by the schema_types_override
        if the datastructure has been initialised with a query, or the relevant table
        schema if the datastructure has been initialised with a table name.

        Args:
            schema: The BitfountSchema either taken from the pod
                or provided by the user when defining a model.
            pod_identifier: The pod identifier(s) on which the
                model will be trained on. Defaults to None.
            datasource: The datasource on which the model
                will be trained on. Defaults to None.
        """
        if self.query:
            # If the datastructure is given a query, then we use the schema override.
            if datasource and isinstance(datasource, DatabaseSource):
                if isinstance(self.query, dict) and pod_identifier:
                    datasource.datastructure_query = self.query[pod_identifier]
                elif isinstance(self.query, str):
                    datasource.datastructure_query = self.query
            table_schema = self._override_schema(
                pod_identifier=pod_identifier, datasource=datasource
            )
        elif self.table:
            # If the datastructure is given a table name, then we get the table schema.
            table_schema = schema.get_table_schema(self.get_table_name(pod_identifier))
        return table_schema

    def _update_datastructure_with_hub_identifiers(
        self, hub_pod_ids: List[str]
    ) -> None:
        """Update the pod_ids with the hub ids, containing username."""
        if self.table and isinstance(self.table, dict):
            self.table = dict(zip(hub_pod_ids, self.table.values()))
        elif (
            self.query
            and isinstance(self.query, dict)
            and self.schema_types_override is not None
        ):
            # By this point if the datastructure has a query,
            # it will include the pod_ids for both schema override and query.
            self.query = dict(zip(hub_pod_ids, self.query.values()))
            self.schema_types_override = cast(
                Mapping[str, SchemaOverrideMapping],
                dict(
                    zip(
                        hub_pod_ids,
                        self.schema_types_override.values(),
                    )
                ),
            )

    class _Schema(MarshmallowSchema):
        table = StrDictField(allow_none=True)
        query = StrDictField(allow_none=True)
        schema_types_override = fields.Dict(allow_none=True)
        target = fields.Raw(required=True)
        ignore_cols = fields.List(fields.Str(), allow_none=True)
        loss_weights_col = fields.Str(allow_none=True)
        multihead_col = fields.Str(allow_none=True)
        ignore_classes_col = fields.Str(allow_none=True)
        batch_transforms = fields.List(
            fields.Dict(
                keys=fields.Str(),
                values=fields.Dict(keys=fields.Str()),
            ),
            allow_none=True,
        )
        dataset_transforms = fields.List(
            fields.Dict(
                keys=fields.Str(),
                values=fields.Dict(keys=fields.Str()),
            ),
            allow_none=True,
        )

        @post_load
        def recreate_datastructure(
            self, data: _JSONDict, **_kwargs: Any
        ) -> DataStructure:
            """Recreates DataStructure."""
            return DataStructure(**data)



"""Tests for the federated averaging protocol."""
from pathlib import Path
import re
from typing import Dict, Tuple, cast
from unittest import mock
from unittest.mock import ANY, AsyncMock, MagicMock, Mock, create_autospec

from _pytest.logging import LogCaptureFixture
import pytest
from pytest import fixture
from pytest_mock import MockerFixture

from bitfount import BitfountModelReference, BitfountSchema
from bitfount.backends.pytorch import PyTorchTabularClassifier
from bitfount.data.datasources.base_source import BaseSource
import bitfount.federated.aggregators
import bitfount.federated.aggregators.aggregator
from bitfount.federated.aggregators.aggregator import Aggregator
from bitfount.federated.aggregators.base import _AggregatorWorkerFactory
from bitfount.federated.aggregators.secure import _InterPodAggregatorWorkerFactory
from bitfount.federated.algorithms.model_algorithms import federated_training
from bitfount.federated.algorithms.model_algorithms.federated_training import (
    FederatedModelTraining,
)
from bitfount.federated.helper import _create_aggregator
import bitfount.federated.protocols.base as protocols
from bitfount.federated.protocols.base import (
    _BaseModellerProtocol,
    _BaseProtocol,
    _BaseProtocolFactory,
    _BaseWorkerProtocol,
)
from bitfount.federated.protocols.model_protocols import federated_averaging
from bitfount.federated.protocols.model_protocols.federated_averaging import (
    FederatedAveraging,
    _FederatedAveragingCompatibleAlgoFactory,
)
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.worker_transport import (
    _InterPodWorkerMailbox,
    _WorkerMailbox,
)
from bitfount.schemas.utils import bf_dump, bf_load
from bitfount.types import _JSONDict
from tests.bitfount.backends.pytorch.models.test_models import assert_vars_equal
from tests.utils import PytestRequest
from tests.utils.helper import backend_test, create_datastructure, unit_test


def mocked_modeller_runner_functions(
    mocker: MockerFixture,
) -> Tuple[Aggregator, AsyncMock, Mock]:
    """Mock functions used in protocol.modeller.run."""
    mocker.patch.object(federated_averaging._ModellerSide, "perform_iterations_checks")

    mock_algorithm_run = Mock()
    mocker.patch.object(
        federated_training._ModellerSide,
        "run",
        mock_algorithm_run,
    )

    mocker.patch.object(
        federated_averaging._ModellerSide,
        "_send_parameters",
    )

    mocker.patch.object(
        federated_averaging._ModellerSide,
        "_receive_parameter_updates",
        return_value=["some params"],
    )

    mocker.patch.object(
        federated_averaging._ModellerSide,
        "get_num_federated_iterations",
        return_value=2,
    )

    mock_get_training_metrics = AsyncMock()
    mocker.patch.object(
        federated_averaging._ModellerSide,
        "_get_training_metrics_updates",
        mock_get_training_metrics,
    )
    mock_get_training_metrics.return_value = {"AUC": "0.7"}

    mock_aggregator_run = Mock()

    mocker.patch.object(
        bitfount.federated.aggregators.aggregator._ModellerSide,
        "run",
        mock_aggregator_run,
    )
    mock_aggregator_run.return_value = {"some_weight": "some_value"}
    mocked_aggregator_factory = Aggregator()

    mock_task_complete = MagicMock()
    mock_task_complete.return_value.set_result(True)
    mocker.patch.object(
        _ModellerMailbox, "send_training_iteration_complete_update", mock_task_complete
    )
    mocker.patch.object(_ModellerMailbox, "send_task_complete_message")
    return mocked_aggregator_factory, mock_get_training_metrics, mock_algorithm_run


def mocked_worker_runner_functions(
    mocker: MockerFixture,
) -> Tuple[Aggregator, AsyncMock]:
    """Mock functions used in protocol.worker.run."""
    mocker.patch.object(federated_averaging._WorkerSide, "perform_iterations_checks")

    mocker.patch.object(
        federated_training._WorkerSide,
        "run",
        return_value=([1, 2, 3], [{"AUC": "0.7"}]),
    )

    mocker.patch.object(
        federated_training._WorkerSide,
        "save_final_parameters",
    )

    mocker.patch.object(
        federated_averaging._WorkerSide,
        "get_num_federated_iterations",
        return_value=2,
    )

    mocker.patch.object(
        federated_averaging._WorkerSide,
        "_send_parameter_update",
    )
    mock_send_training_metrics = AsyncMock()
    mocker.patch.object(
        federated_averaging._WorkerSide,
        "_send_training_metrics",
        mock_send_training_metrics,
    )

    mock_aggregator_run = AsyncMock()

    mocker.patch.object(
        bitfount.federated.aggregators.aggregator._WorkerSide,
        "run",
        mock_aggregator_run,
    )
    mocked_aggregator_factory = Aggregator()

    mock_task_complete = AsyncMock(return_value=True)
    mocker.patch.object(
        _WorkerMailbox, "get_training_iteration_complete_update", mock_task_complete
    )
    mocker.patch.object(_WorkerMailbox, "get_task_complete_update", mock_task_complete)
    return mocked_aggregator_factory, mock_send_training_metrics


class TestFederatedAveraging:
    """Test Federated Averaging protocol."""

    @fixture(scope="function", params=[None, "/mock/file"])
    def federated_algorithm(self, request: PytestRequest) -> FederatedModelTraining:
        """Returns federated algorithm."""
        return FederatedModelTraining(model=Mock(), pretrained_file=request.param)

    @fixture
    def mock_federated_algorithm(self) -> Mock:
        """Returns a mock algorithm compatible with FederatedAveraging."""
        mock_algorithm: Mock = create_autospec(
            _FederatedAveragingCompatibleAlgoFactory, instance=True
        )
        return mock_algorithm

    @fixture
    def mock_aggregator(self) -> Mock:
        """Returns mock aggregator."""
        mock_aggregator: Mock = create_autospec(Aggregator, instance=True)
        return mock_aggregator

    @fixture
    def mocked_modeller_runner_fixture(
        self,
        mocker: MockerFixture,
    ) -> Tuple[Aggregator, AsyncMock, Mock]:
        """Fixture for getting the necessary mocks for modeller protocol runner."""
        return mocked_modeller_runner_functions(mocker)

    @fixture
    def mocked_worker_runner_fixture(
        self, mocker: MockerFixture
    ) -> Tuple[Aggregator, AsyncMock]:
        """Fixture for getting the necessary mocks for worker protocol runner."""
        return mocked_worker_runner_functions(mocker)

    @unit_test
    def test_algorithm_not_compatible_raises_type_error(
        self,
        mock_aggregator: Mock,
    ) -> None:
        """Check that TypeError is raised if algorithm is not compatible."""
        mock_algorithm: Mock = Mock(spec_set=["__name__"])
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"The {FederatedAveraging.__name__} protocol does "
                f"not support the {type(mock_algorithm).__name__} algorithm."
            ),
        ):
            FederatedAveraging(
                algorithm=mock_algorithm,
                aggregator=mock_aggregator,
                steps_between_parameter_updates=2,
            )

    @unit_test
    def test_modeller(
        self,
        federated_algorithm: FederatedModelTraining,
        mock_aggregator: Mock,
        mock_modeller_mailbox: Mock,
    ) -> None:
        """Test modeller method."""
        protocol_factory = FederatedAveraging(
            algorithm=federated_algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=2,
        )
        protocol = protocol_factory.modeller(mailbox=mock_modeller_mailbox)

        for type_ in [
            _BaseProtocol,
            _BaseModellerProtocol,
            federated_averaging._ModellerSide,
        ]:
            assert isinstance(protocol, type_)

    @unit_test
    def test_worker(
        self,
        federated_algorithm: FederatedModelTraining,
        mock_aggregator: Mock,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
    ) -> None:
        """Test worker method."""
        protocol_factory = FederatedAveraging(
            algorithm=federated_algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=2,
        )
        protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)

        for type_ in [
            _BaseProtocol,
            _BaseWorkerProtocol,
            federated_averaging._WorkerSide,
        ]:
            assert isinstance(protocol, type_)

    @unit_test
    def test_worker_with_different_aggregator_types(
        self,
        mock_federated_algorithm: Mock,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Test worker method with different aggregator types."""
        # Mock out WorkerSide constructor
        mock_worker_side_cls = mocker.patch(
            "bitfount.federated.protocols.model_protocols.federated_averaging._WorkerSide",
            autospec=True,
        )

        # Test with an instance of AggregatorWorkerFactory
        mock_aggregator: Mock = create_autospec(_AggregatorWorkerFactory, instance=True)
        protocol_factory = FederatedAveraging(
            algorithm=mock_federated_algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=2,
        )
        protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
        # Check WorkerSide constructed as expected
        assert protocol == mock_worker_side_cls.return_value
        mock_worker_side_cls.assert_called_once_with(
            algorithm=mock_federated_algorithm.worker.return_value,
            aggregator=mock_aggregator.worker.return_value,
            steps_between_parameter_updates=2,
            epochs_between_parameter_updates=None,
            auto_eval=ANY,
            mailbox=mock_worker_mailbox,
        )
        # Check aggregator.worker() called as expected
        mock_aggregator.worker.assert_called_once_with()

        # Test with an instance of InterPodAggregatorWorkerFactory
        mock_worker_side_cls.reset_mock()
        mock_aggregator = create_autospec(
            _InterPodAggregatorWorkerFactory, instance=True
        )
        mock_interpod_worker_mailbox = create_autospec(
            _InterPodWorkerMailbox, instance=True
        )
        protocol_factory = FederatedAveraging(
            algorithm=mock_federated_algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=2,
        )
        protocol = protocol_factory.worker(
            mailbox=mock_interpod_worker_mailbox, hub=mock_hub
        )
        # Check WorkerSide constructed as expected
        assert protocol == mock_worker_side_cls.return_value
        mock_worker_side_cls.assert_called_once_with(
            algorithm=mock_federated_algorithm.worker.return_value,
            aggregator=mock_aggregator.worker.return_value,
            steps_between_parameter_updates=2,
            epochs_between_parameter_updates=None,
            auto_eval=ANY,
            mailbox=mock_interpod_worker_mailbox,
        )
        # Check aggregator.worker() called as expected
        mock_aggregator.worker.assert_called_once_with(
            mailbox=mock_interpod_worker_mailbox
        )

        # Test with an unknown type of aggregator instance
        mock_aggregator = Mock()
        protocol_factory = FederatedAveraging(
            algorithm=mock_federated_algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=2,
        )
        with pytest.raises(TypeError, match="Unrecognised aggregator factory"):
            protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)

    @unit_test
    def test_helper_run_method_without_algorithm_raises_type_error(
        self,
    ) -> None:
        """Tests helper run method without algorithm raises TypeError."""
        with pytest.raises(TypeError):
            FederatedAveraging()  # type: ignore[call-arg] # Reason: this is what we are testing for # noqa: B950

    @unit_test
    def test_helper_run_method_with_algorithm(
        self, mock_hub: Mock, mocker: MockerFixture
    ) -> None:
        """Tests helper run method with algorithm."""
        federated_algorithm_with_model = FederatedModelTraining(model=Mock())
        mock_run_protocol = mocker.patch.object(_BaseProtocolFactory, "run")
        mock_create_aggregator = mocker.patch(
            "bitfount.federated.protocols.model_protocols.federated_averaging._create_aggregator"
        )

        protocol = FederatedAveraging(algorithm=federated_algorithm_with_model)

        protocol.run(
            pod_identifiers=["bitfount/fake", "bitfount/fake2"],
            hub=mock_hub,
            private_key_or_file=Path("fake.testkey"),
        )

        mock_run_protocol.assert_called_once()
        mock_create_aggregator.assert_called()

    @unit_test
    def test_helper_run_method_with_model(
        self, mock_hub: Mock, mocker: MockerFixture
    ) -> None:
        """Tests helper run method with a model."""
        federated_model = Mock()

        mock_run_protocol = mocker.patch.object(_BaseProtocolFactory, "run")
        mock_create_aggregator = mocker.patch(
            "bitfount.federated.protocols.model_protocols.federated_averaging._create_aggregator"
        )
        protocol = FederatedAveraging(
            algorithm=FederatedModelTraining(model=federated_model)
        )

        protocol.run(
            pod_identifiers=["bitfount/fake", "bitfount/fake2"],
            hub=mock_hub,
            private_key_or_file=Path("fake.testkey"),
        )

        mock_run_protocol.assert_called_once()
        mock_create_aggregator.assert_called()

    @unit_test
    def test_helper_run_method_with_model_and_algorithm(
        self, caplog: LogCaptureFixture, mock_hub: Mock, mocker: MockerFixture
    ) -> None:
        """Tests helper run method with an algorithm and a model.

        This tests that the run method will still run but that it just issues a warning
        regarding the extra model argument.
        """
        model = Mock()
        federated_algorithm_with_model = FederatedModelTraining(model=Mock())

        mock_run_protocol = mocker.patch.object(_BaseProtocolFactory, "run")
        mock_create_aggregator = mocker.patch(
            "bitfount.federated.protocols.model_protocols.federated_averaging._create_aggregator",
        )

        protocol = FederatedAveraging(
            model=model, algorithm=federated_algorithm_with_model
        )
        protocol.run(
            pod_identifiers=["bitfount/fake", "bitfount/fake2"],
            hub=mock_hub,
            private_key_or_file=Path("fake.testkey"),
        )

        mock_run_protocol.assert_called_once()
        mock_create_aggregator.assert_called()

        model.assert_not_called()
        model.backend_tensor_shim.assert_not_called()
        assert (
            caplog.records[0].msg
            == "Ignoring provided model. Algorithm already has a model."
        )

    @unit_test
    async def test_worker_auto_eval_true(
        self,
        federated_algorithm: FederatedModelTraining,
        mock_aggregator: Mock,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        mocked_worker_runner_fixture: Tuple[Aggregator, AsyncMock],
    ) -> None:
        """Test worker method, validation metrics sent to modeller."""
        (
            mocked_aggregator_factory,
            mock_send_training_metrics,
        ) = mocked_worker_runner_fixture

        protocol_factory = FederatedAveraging(
            algorithm=federated_algorithm,
            aggregator=mocked_aggregator_factory,
            steps_between_parameter_updates=2,
            auto_eval=True,
        )
        protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
        await protocol.run(datasource=create_autospec(BaseSource, instance=True))
        mock_send_training_metrics.assert_called_once()

    @unit_test
    async def test_worker_auto_eval_false(
        self,
        federated_algorithm: FederatedModelTraining,
        mock_aggregator: Mock,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
        mocked_worker_runner_fixture: Tuple[Aggregator, AsyncMock],
    ) -> None:
        """Test worker method no validation metrics sent."""
        (
            mocked_aggregator_factory,
            mock_send_training_metrics,
        ) = mocked_worker_runner_fixture

        protocol_factory = FederatedAveraging(
            algorithm=federated_algorithm,
            aggregator=mocked_aggregator_factory,
            steps_between_parameter_updates=2,
            auto_eval=False,
        )
        protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
        await protocol.run(datasource=create_autospec(BaseSource, instance=True))
        mock_send_training_metrics.assert_not_called()

    @unit_test
    def test_worker_raises_exception_interpod_communication_incorrect(
        self,
        mock_hub: Mock,
        mock_worker_mailbox: Mock,
    ) -> None:
        """Tests exception raised if interpod mailbox needed but not provided."""
        protocol_factory = FederatedAveraging(
            algorithm=Mock(),
            aggregator=create_autospec(_InterPodAggregatorWorkerFactory, instance=True),
        )

        with pytest.raises(
            TypeError,
            match=re.escape(
                "Inter-pod aggregators require an inter-pod worker mailbox."
            ),
        ):
            protocol_factory.worker(mock_worker_mailbox, mock_hub)

    @unit_test
    async def test_modeller_auto_eval_true(
        self,
        federated_algorithm: FederatedModelTraining,
        mock_aggregator: Mock,
        mock_hub: Mock,
        mock_modeller_mailbox: Mock,
        mocked_modeller_runner_fixture: Tuple[Aggregator, AsyncMock, Mock],
    ) -> None:
        """Test worker method no validation metrics sent."""
        (
            mocked_aggregator_factory,
            mock_get_training_metrics,
            mock_algorithm_run,
        ) = mocked_modeller_runner_fixture

        protocol_factory = FederatedAveraging(
            algorithm=federated_algorithm,
            aggregator=mocked_aggregator_factory,
            steps_between_parameter_updates=1,
        )
        protocol = protocol_factory.modeller(
            mailbox=mock_modeller_mailbox, hub=mock_hub
        )
        await protocol.run()
        mock_get_training_metrics.assert_called()
        assert protocol.validation_results == [{"AUC": "0.7"}, {"AUC": "0.7"}]
        algorithm_mock_calls = [
            mock.call(update=None),
            mock.call(
                update={"some_weight": "some_value"}, validation_metrics={"AUC": "0.7"}
            ),
            mock.call(
                update={"some_weight": "some_value"}, validation_metrics={"AUC": "0.7"}
            ),
        ]
        assert mock_algorithm_run.mock_calls == algorithm_mock_calls

    @unit_test
    async def test_modeller_auto_eval_false(
        self,
        federated_algorithm: FederatedModelTraining,
        mock_aggregator: Mock,
        mock_hub: Mock,
        mock_modeller_mailbox: Mock,
        mocked_modeller_runner_fixture: Tuple[Aggregator, AsyncMock, Mock],
    ) -> None:
        """Test worker method no validation metrics sent."""
        (
            mocked_aggregator_factory,
            mock_get_training_metrics,
            mock_algorithm_run,
        ) = mocked_modeller_runner_fixture
        protocol_factory = FederatedAveraging(
            algorithm=federated_algorithm,
            aggregator=mocked_aggregator_factory,
            steps_between_parameter_updates=2,
            auto_eval=False,
        )
        protocol = protocol_factory.modeller(
            mailbox=mock_modeller_mailbox, hub=mock_hub
        )
        await protocol.run()
        mock_get_training_metrics.assert_not_called()
        assert protocol.validation_results == []
        algorithm_mock_calls = [
            mock.call(update=None),
            mock.call(update={"some_weight": "some_value"}),
            mock.call(update={"some_weight": "some_value"}),
        ]
        assert mock_algorithm_run.mock_calls == algorithm_mock_calls

    @unit_test
    @pytest.mark.parametrize(
        "iterations, model_reference",
        [
            ({"epochs": 1, "steps": None}, True),
            ({"epochs": None, "steps": 1}, True),
            ({"epochs": 1, "steps": None}, False),
            ({"epochs": None, "steps": 1}, False),
        ],
    )
    async def test_federated_averaging_protocol_no_iter(
        self, iterations: Dict[str, int], mock_aggregator: Mock, model_reference: bool
    ) -> None:
        """Tests FederatedAveraging with no iteration provided.

        Tests that iterations is set to a default value if no value
        is provided.
        """
        if model_reference:
            model = create_autospec(BitfountModelReference)
            model.hyperparameters = iterations
        else:
            model = Mock(**iterations)
        algorithm = FederatedModelTraining(model=model)

        protocol = FederatedAveraging(
            algorithm=algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=None,
            epochs_between_parameter_updates=None,
            auto_eval=False,
        )

        if iterations.get("steps"):
            assert protocol.steps_between_parameter_updates == 1
            assert protocol.epochs_between_parameter_updates is None
        else:
            assert protocol.epochs_between_parameter_updates == 1
            assert protocol.steps_between_parameter_updates is None

    @unit_test
    @pytest.mark.parametrize("secure_aggregation", [True, False])
    async def test_federated_averaging_protocol_successfully(
        self, secure_aggregation: bool
    ) -> None:
        """Tests that FederatedAveraging works as expected."""
        model = Mock(epochs=1)

        algorithm = FederatedModelTraining(model=model)

        aggregator = _create_aggregator(secure_aggregation=secure_aggregation)
        protocol = FederatedAveraging(
            algorithm=algorithm,
            aggregator=aggregator,
            steps_between_parameter_updates=None,
            epochs_between_parameter_updates=2,
            auto_eval=False,
        )
        assert isinstance(protocol, FederatedAveraging)
        assert protocol.epochs_between_parameter_updates == 2
        assert protocol.steps_between_parameter_updates is None

    @unit_test
    def test_protocol_datastructure_error(self, mock_hub: Mock) -> None:
        """Tests that protocol raises `DataStructure` error."""
        model = Mock(epochs=1, datastructure=create_datastructure())

        algorithm = FederatedModelTraining(model=model)

        protocol = FederatedAveraging(
            algorithm=algorithm,
            steps_between_parameter_updates=None,
            epochs_between_parameter_updates=2,
            auto_eval=False,
        )
        with pytest.raises(ValueError):
            protocol.run(pod_identifiers=["pod1", "pod2"], hub=mock_hub)


@backend_test
@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for FederatedAveraging protocol."""

    def test_serialization_model_algorithm(self) -> None:
        """Test Marshmallow Serialization for FederatedAveraging protocol."""
        model = PyTorchTabularClassifier(
            datastructure=create_datastructure(),
            schema=BitfountSchema(),
            epochs=2,
        )
        algorithm_factory = FederatedModelTraining(model=model)
        fed_avg = FederatedAveraging(
            algorithm=algorithm_factory, aggregator=Aggregator()
        )

        dumped = bf_dump(fed_avg)
        loaded = bf_load(dumped, protocols.registry)

        assert vars(fed_avg).keys() == vars(loaded).keys()
        assert fed_avg.algorithm.class_name == loaded.algorithm.class_name
        assert fed_avg.aggregator.class_name == loaded.aggregator.class_name

        assert_vars_equal(vars(fed_avg.algorithm.model), vars(loaded.algorithm.model))

    def test_serialization_model_algorithm_from_protocol(self) -> None:
        """Test Marshmallow Serialization for FederatedAveraging protocol."""
        model = PyTorchTabularClassifier(
            datastructure=create_datastructure(),
            schema=BitfountSchema(),
            epochs=2,
        )
        algorithm_factory = FederatedModelTraining(model=model)
        fed_avg = FederatedAveraging(
            algorithm=algorithm_factory, aggregator=Aggregator()
        )

        dumped = fed_avg.dump()
        loaded = bf_load(cast(_JSONDict, dumped), protocols.registry)

        assert vars(fed_avg).keys() == vars(loaded).keys()
        assert fed_avg.algorithm.class_name == loaded.algorithm.class_name
        assert fed_avg.aggregator.class_name == loaded.aggregator.class_name

        assert_vars_equal(vars(fed_avg.algorithm.model), vars(loaded.algorithm.model))



"""Test metrics from bitfount.metrics."""
from dataclasses import dataclass
from typing import Optional, Union, cast

from _pytest.logging import LogCaptureFixture
import matplotlib
import numpy as np
import pandas as pd
import pytest
from pytest_mock import MockerFixture
from sklearn.metrics import recall_score

from bitfount.metrics import (
    BINARY_CLASSIFICATION_METRICS,
    MULTICLASS_CLASSIFICATION_METRICS,
    MULTILABEL_CLASSIFICATION_METRICS,
    REGRESSION_METRICS,
    SEGMENTATION_METRICS,
    ClassificationMetric,
    MetricCollection,
    MetricsProblem,
    _binary_dice_coefficient,
    _stats_score,
    dice_score,
    iou,
    multiclass_dice_coefficients,
    sum_fpr_curve,
)
from bitfount.models.base_models import ClassifierMixIn, RegressorMixIn
from tests.utils.helper import integration_test, unit_test


class TestMetricCollection:
    """Tests Metrics class explicitly where necessitated.

    Currently already being partially tested in `test_models.py`
    """

    @unit_test
    def test_value_error_if_no_metrics_or_problem_provided(self) -> None:
        """Tests that a ValueError is raised in the constructor.

        This happens if neither `problem` nor `metrics` are provided.
        """
        with pytest.raises(ValueError):
            MetricCollection()

    @unit_test
    @pytest.mark.parametrize(
        "classification, multilabel, classes, problem",
        [
            (True, False, 2, MetricsProblem.BINARY_CLASSIFICATION),
            (True, False, 3, MetricsProblem.MULTICLASS_CLASSIFICATION),
            (True, True, 2, MetricsProblem.MULTILABEL_CLASSIFICATION),
            (False, None, None, MetricsProblem.REGRESSION),
        ],
    )
    def test_create_instance_from_model(
        self,
        classes: Optional[int],
        classification: bool,
        multilabel: bool,
        problem: MetricsProblem,
    ) -> None:
        """Test Metric Collection created from Model.

        Tests that MetricCollection creates the right problem from the Model.
        """
        if isinstance(classes, int):
            n_classes = classes

        @dataclass
        class ClassifierModel(ClassifierMixIn):
            multilabel: bool
            n_classes: int

        class RegressorModel(RegressorMixIn):
            pass

        model: Union[ClassifierModel, RegressorModel]
        if classification:
            model = ClassifierModel(multilabel, n_classes)
        else:
            model = RegressorModel()

        m = MetricCollection.create_from_model(model)  # type: ignore[arg-type] # reason: used only for testing # noqa: B950
        assert m.problem == problem

    @unit_test
    def test_create_instance_from_unrecognised_model(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Test Metric Collection created from an unrecognised model."""
        MetricCollection.create_from_model(model="model", metrics=REGRESSION_METRICS)  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950
        log_record = caplog.records[0]
        assert log_record.levelname == "WARNING"
        assert (
            log_record.message
            == "Metrics problem type can't be determined. Leaving empty."
        )

    @integration_test
    def test_binary_classification(self) -> None:
        """Test metrics in the binary class setting."""
        y_true = np.array([0, 1, 1, 0, 1, 1])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])

        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
        m.compute(y_true, y_prob)

        assert isinstance(m.results, dict)
        assert len(m.results) == len(BINARY_CLASSIFICATION_METRICS)
        assert cast(float, m.optimal_threshold) <= 1.0
        assert cast(float, m.optimal_threshold) >= 0.0
        for _, v in m.results.items():
            # assert 1 number before the decimal place and a maximum of 4 after
            assert len(str(v)) <= 6
            assert v >= 0.0
            assert v <= 1.0

    @integration_test
    def test_binary_classification_not_enough_data(self) -> None:
        """Test metrics in the binary class setting with not enough data."""
        y_true = np.array([0, 0])
        y_prob = np.array([0.1, 0.3])

        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
        m.compute(y_true, y_prob)

        assert isinstance(m.results, dict)
        assert len(m.results) == len(BINARY_CLASSIFICATION_METRICS)
        assert m.results["AUC"] == np.inf

    @integration_test
    def test_multiclass_classification(self) -> None:
        """Test metrics in the multiclass setting (>2 classes)."""
        y_true = np.array([0, 1, 2, 0, 1])
        y_prob = np.array(
            [
                [0.5, 0.4, 0.1],
                [0.9, 0.1, 0.0],
                [0.0, 0.5, 0.5],
                [0.3, 0.3, 0.4],
                [0.2, 0.6, 0.2],
            ]
        )

        m = MetricCollection(problem=MetricsProblem.MULTICLASS_CLASSIFICATION)
        m.compute(y_true, y_prob)
        assert isinstance(m.results, dict)
        assert len(m.results) == len(MULTICLASS_CLASSIFICATION_METRICS)
        for _, v in m.results.items():
            # assert 1 number before the decimal place and a maximum of 4 after
            assert len(str(v)) <= 6
            assert v >= 0.0
            assert v <= 1.0

    @integration_test
    def test_multilabel_classification(self) -> None:
        """Test metrics in the multilabel setting."""
        y_true = np.array([[1, 1, 0], [1, 0, 0], [0, 1, 0], [0, 1, 0], [1, 1, 0]])
        y_prob = np.array(
            [
                [0.5, 0.4, 0.1],
                [0.9, 0.1, 0.0],
                [0.0, 0.5, 0.5],
                [0.3, 0.3, 0.4],
                [0.2, 0.6, 0.2],
            ]
        )

        m = MetricCollection(problem=MetricsProblem.MULTILABEL_CLASSIFICATION)
        m.compute(y_true, y_prob)
        assert isinstance(m.results, dict)
        assert len(m.results) == len(MULTILABEL_CLASSIFICATION_METRICS)
        assert cast(float, m.optimal_threshold) <= 1.0
        assert cast(float, m.optimal_threshold) >= 0.0
        for _, v in m.results.items():
            # assert 1 number before the decimal place and a maximum of 4 after
            assert len(str(v)) <= 6
            assert v >= 0.0
            assert v <= 1.0

    @integration_test
    def test_regression(self) -> None:
        """Test metrics in the regression setting."""
        y_true = np.array([0.2, 1, 0.5, 0, 1, 0.55])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])

        m = MetricCollection(problem=MetricsProblem.REGRESSION)
        m.compute(y_true, y_prob)

        assert isinstance(m.results, dict)
        assert len(m.results) == len(REGRESSION_METRICS)
        assert m.optimal_threshold is None

    @integration_test
    def test_optimise_chosen_metric(self) -> None:
        """Optimise chosen metric in binary classification."""
        y_true = np.array([0, 1, 1, 0, 1, 1])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])

        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
        m.compute(y_true, y_prob, "Recall")
        assert m.optimal_threshold == 0.0

    @unit_test
    def test_optimise_chosen_metric_does_not_exist(self) -> None:
        """Test optimisation for classification.

        Chosen metric doesn not exist raises ValueError.
        """
        y_true = np.array([0, 1, 1, 0, 1, 1])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])

        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
        with pytest.raises(ValueError):
            m.compute(y_true, y_prob, "Blah")

    @unit_test
    def test_get_results_df(self) -> None:
        """Tests get_results_df method."""
        y_true = np.array([0, 1, 1, 0, 1, 1])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])

        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
        m.compute(y_true, y_prob)
        results_df = m.get_results_df()

        assert isinstance(results_df, pd.DataFrame)
        for index, row in results_df.iterrows():
            assert row["value"] == m.results[cast(str, index)]

    @integration_test
    def test_threshold(self) -> None:
        """Supply the threshold rather than optimising it."""
        y_true = np.array([0, 1, 1, 0, 1, 1])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])

        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
        m.compute(y_true, y_prob, threshold=0.5)
        assert m.results["Recall"] == 0.75
        assert m.optimal_threshold is None

    @unit_test
    def test_view_thresholds(self) -> None:
        """Tests that chart displaying thresholds can be created properly.

        Including the underlying data.
        """
        y_true = np.array([0, 1, 1, 0, 1, 1])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])

        m = MetricCollection(problem=MetricsProblem.BINARY_CLASSIFICATION)
        m.compute(y_true, y_prob)

        assert len(m.thresholds) == 201  # num of thresholds (0-100 at 0.5 increments)
        assert isinstance(m.threshold_metrics, dict)
        for k, v in m.threshold_metrics.items():
            assert k in m.metrics
            assert len(v) == len(m.thresholds)

        fig = m.view_thresholds()
        assert isinstance(fig, matplotlib.figure.Figure)

    @integration_test
    def test_custom_metric(self) -> None:
        """Test custom metric can be supplied."""
        y_true = np.array([0, 1, 1, 0, 1, 1])
        y_prob = np.array([0.1, 0.9, 0.8, 0.3, 0.2, 0.5])
        custom_metric = ClassificationMetric(recall_score, False)
        m = MetricCollection(
            problem=MetricsProblem.BINARY_CLASSIFICATION,
            metrics={"custom_metric": custom_metric},
        )
        m.compute(y_true, y_prob, metric_to_optimise="custom_metric")
        assert isinstance(m.results, dict)
        assert len(m.results) == 1
        assert m.optimal_threshold is not None

    @unit_test
    def test_correct__get_metrics_segmentation_problem(self) -> None:
        """Test that segmentation metrics are loaded correctly."""
        m = MetricCollection(problem=MetricsProblem.SEGMENTATION)
        assert "IoU" in m.metrics
        assert "DiceCoefficients" in m.metrics
        assert "DiceScore" in m.metrics

    @unit_test
    def test_segmentation_metrics(self) -> None:
        """Test metrics in segmentation setting."""
        y_true = np.array([[1, 1, 0], [1, 0, 0], [0, 1, 0], [0, 1, 0], [1, 1, 0]])
        y_prob = np.array(
            [
                [0.5, 0.4, 0.1],
                [0.9, 0.1, 0.0],
                [0.0, 0.5, 0.5],
                [0.3, 0.3, 0.4],
                [0.2, 0.6, 0.2],
            ]
        )

        m = MetricCollection(problem=MetricsProblem.SEGMENTATION)
        m.compute(y_true, y_prob)
        assert isinstance(m.results, dict)
        assert len(m.results) == len(SEGMENTATION_METRICS)
        assert m.optimal_threshold is None

    @unit_test
    def test_get_metrics_value_error(self) -> None:
        """Test that get_metrics raises ValueError if metric is not found."""
        with pytest.raises(ValueError, match="Problem type not recognised."):
            MetricCollection._get_metrics("not a problem")  # type: ignore[arg-type] # Reason: purpose of test # noqa: B950


@unit_test
class TestSumFPRCurve:
    """Tests sum_fpr_curve() function."""

    def test_one_entry(self) -> None:
        """Tests functionality when only one entry.

        Check that when we have one entry correct, one wrong it gets placed in
        the right places.
        """
        test_target = [1.0, 0.0]
        test_pred = [0.89, 0.32]
        test_var_to_sum = [1.0, -2.0]
        fpr, totals = sum_fpr_curve(
            test_target, test_pred, test_var_to_sum, granularity=10
        )
        # At threshold of 0.3 we start accepting the bad one
        assert fpr == [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0]
        # At threshold of 0.8 we start accepting the good one
        assert totals == [-1, -1, -1, -1, 1, 1, 1, 1, 1, 0, 0]


@unit_test
def test__binary_dice_coefficients() -> None:
    """Tests the binary dice coefficients.

    Computes the expected value as is 2 * the area of intersection divided by
    the total number of pixels in both images.

    """
    y_true = np.array([[1, 1, 0], [1, 0, 0], [0, 1, 0], [0, 1, 0], [1, 1, 0]])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    # Flatten the target and predictions arrays.
    y_true_f = y_true.flatten()
    y_pred_f = y_prob.flatten()
    # Compute the intersection between predictions and targets
    intersection = (y_true_f * y_pred_f).sum()
    # Get the expected result
    expected_result = (2.0 * intersection) / (y_true_f.sum() + y_pred_f.sum() + 0.0001)
    # Get the result returned from the `_binary_dice_coefficient` function
    res = _binary_dice_coefficient(y_true, y_prob)
    # Make sure the resultis the same
    assert expected_result == res


@unit_test
def test_multiclass_dice_coeff(mocker: MockerFixture) -> None:
    """Tests the multiclass dice coefficients."""
    y_true = np.array([[1, 1, 0], [1, 0, 0], [0, 1, 0], [0, 1, 0], [1, 1, 0]])
    y_prob = np.random.uniform(0, 1, size=(4, 3, 3, 2))
    mock_binary_coeff = mocker.patch(
        "bitfount.metrics._binary_dice_coefficient", return_value=0.6
    )
    res = np.round(multiclass_dice_coefficients(y_true, y_prob), 1)
    mock_binary_coeff.assert_called()
    assert res == 0.4


@unit_test
def test__stats_score() -> None:
    """Tests the stats score function."""
    y_true = np.array([0, 1, 1, 0, 1])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    tp, fp, tn, fn = _stats_score(y_true, y_prob)
    assert tp == 2
    assert fp == 0
    assert tn == 2
    assert fn == 1


@unit_test
def test_dice_score(mocker: MockerFixture) -> None:
    """Tests the dice score function."""
    y_true = np.array([0, 1, 1, 0, 1])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    mock_stats_score = mocker.patch(
        "bitfount.metrics._stats_score", return_value=(2, 0, 0, 0)
    )
    score = dice_score(y_true, y_prob)
    mock_stats_score.assert_called()
    assert score == 1.0 / 3


@unit_test
def test_dice_score_no_foreground(mocker: MockerFixture) -> None:
    """Tests dice score zero no foreground class."""
    y_true = np.array([3, 4, 4, 4, 5])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    mock_stats_score = mocker.patch(
        "bitfount.metrics._stats_score", return_value=(2, 0, 0, 0)
    )
    score = dice_score(y_true, y_prob)
    mock_stats_score.assert_called()
    assert score == 0.0


@unit_test
def test_dice_score_denom_zero(mocker: MockerFixture) -> None:
    """Tests dice score zero when denom is zero."""
    y_true = np.array([0, 1, 1, 0, 1])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    mock_stats_score = mocker.patch(
        "bitfount.metrics._stats_score", return_value=(0, 0, 0, 0)
    )
    score = dice_score(y_true, y_prob)
    mock_stats_score.assert_called()
    assert score == 0.0


@unit_test
def test_iou(mocker: MockerFixture) -> None:
    """Tests iou score."""
    y_true = np.array([0, 1, 1, 0, 1])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    mock_stats_score = mocker.patch(
        "bitfount.metrics._stats_score", return_value=(2, 0, 0, 0)
    )
    score = iou(y_true, y_prob)
    mock_stats_score.assert_called()
    assert score == 1.0 / 3


@unit_test
def test_iou_no_foreground_class(mocker: MockerFixture) -> None:
    """Tests iou score zero no foreground class."""
    y_true = np.array([3, 4, 4, 4, 5])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    mock_stats_score = mocker.patch(
        "bitfount.metrics._stats_score", return_value=(2, 0, 0, 0)
    )
    score = iou(y_true, y_prob)
    mock_stats_score.assert_called()
    assert score == 0.0


@unit_test
def test_iou_denom_zero(mocker: MockerFixture) -> None:
    """Tests iou score zero when denom is zero."""
    y_true = np.array([0, 1, 1, 0, 1])
    y_prob = np.array(
        [
            [0.5, 0.4, 0.1],
            [0.9, 0.1, 0.0],
            [0.0, 0.5, 0.5],
            [0.3, 0.3, 0.4],
            [0.2, 0.6, 0.2],
        ]
    )
    mock_stats_score = mocker.patch(
        "bitfount.metrics._stats_score", return_value=(0, 0, 0, 0)
    )
    score = iou(y_true, y_prob)
    mock_stats_score.assert_called()
    assert score == 0.0
